# -*- coding: utf-8 -*- #
from builtins import ValueError

from pyravendb.subscriptions.data import SubscriptionState
from pyravendb.custom_exceptions import exceptions
from pyravendb.tools.utils import Utils
from abc import abstractmethod
from datetime import timedelta
import collections
import string
import random
import logging

logging.basicConfig(filename='responses.log', level=logging.DEBUG)
log = logging.getLogger()


# Todo update the commands
class RavenCommand(object):
    def __init__(self, url=None, method=None, data=None, files=None, headers=None, is_read_request=False,
                 use_stream=False):
        self.url = url
        self.method = method
        self.data = data
        self.headers = headers
        if self.headers is None:
            self.headers = {}
        self.__raven_command = True
        self.is_read_request = is_read_request
        self.use_stream = use_stream
        self.failed_nodes = {}
        self.timeout = None
        self.requested_node = None
        self.files = files

    @abstractmethod
    def create_request(self, server_node):
        raise NotImplementedError

    @abstractmethod
    def set_response(self, response):
        raise NotImplementedError

    @property
    def raven_command(self):
        return self.__raven_command

    def is_failed_with_node(self, node):
        return len(self.failed_nodes) > 0 and node in self.failed_nodes


class GetDocumentCommand(RavenCommand):
    def __init__(self, key_or_keys, includes=None, metadata_only=False):
        """
        @param key_or_keys: the key of the documents you want to retrieve (key can be a list of ids)
        :type str or list
        @param includes: array of paths in documents in which server should look for a 'referenced' document
        :type list
        @param metadata_only: specifies if only document metadata should be returned
        :type bool
        @return: A list of the id or ids we looked for (if they exists)
        :rtype: dict
        """
        super(GetDocumentCommand, self).__init__(method="GET", is_read_request=True)
        self.key_or_keys = key_or_keys
        self.includes = includes
        self.metadata_only = metadata_only

    def create_request(self, server_node):
        if self.key_or_keys is None:
            raise ValueError("None Key is not valid")
        path = "docs?"
        if self.includes:
            path += "".join("&include=" + Utils.quote_key(item) for item in self.includes)
        # make get method handle a multi document requests in a single request
        if isinstance(self.key_or_keys, list):
            key_or_keys = collections.OrderedDict.fromkeys(self.key_or_keys)
            if self.metadata_only:
                path += "&metadataOnly=true"

            # If it is too big, we drop to POST (note that means that we can't use the HTTP cache any longer)
            if (sum(len(x) for x in key_or_keys)) > 1024:
                self.method = "POST"
                self.data = list(key_or_keys)
            else:
                path += "".join("&id=" + Utils.quote_key(item) for item in key_or_keys)

        else:
            path += "&id={0}".format(Utils.quote_key(self.key_or_keys))

        self.url = "{0}/databases/{1}/{2}".format(server_node.url, server_node.database, path)

    def set_response(self, response):
        if response is None:
            return None
        try:
            response = response.json()
            if "Error" in response:
                raise exceptions.ErrorResponseException(response["Error"])
        except ValueError:
            raise exceptions.ErrorResponseException(
                "Failed to load document from the database please check the connection to the server")

        return response


class DeleteDocumentCommand(RavenCommand):
    def __init__(self, key, change_vector=None):
        super(DeleteDocumentCommand, self).__init__(method="DELETE")
        self.key = key
        self.change_vector = change_vector

    def create_request(self, server_node):
        if self.key is None:
            raise ValueError("None Key is not valid")
        if not isinstance(self.key, str):
            raise ValueError("key must be {0}".format(type("")))

        if self.change_vector is not None:
            self.headers = {"If-Match": "\"{0}\"".format(self.change_vector)}

        self.url = "{0}/databases/{1}/docs?id={2}".format(server_node.url, server_node.database,
                                                          Utils.quote_key(self.key))

    def set_response(self, response):
        if response is None:
            log.info("Couldn't find The Document {0}".format(self.key))
        elif response.status_code != 204:
            response = response.json()
            raise exceptions.ErrorResponseException(response["Error"])


class PutDocumentCommand(RavenCommand):
    def __init__(self, key, document, change_vector=None):
        """
        @param str key: unique key under which document will be stored
        @param dict document: document data
        @param str change_vector: current document change_vector, used for concurrency checks (null to skip check)
        @return: json file
        :rtype: dict
        """
        super(PutDocumentCommand, self).__init__(method="PUT")
        self.key = key
        self.document = document
        self.change_vector = change_vector

    def create_request(self, server_node):
        if self.document is None:
            self.document = {}
        if self.change_vector is not None:
            self.headers = {"If-Match": "\"{0}\"".format(self.change_vector)}
        if not isinstance(self.key, str):
            raise ValueError("key must be {0}".format(type("")))
        if not isinstance(self.document, dict):
            raise ValueError("document and metadata must be dict")

        self.data = self.document
        self.url = "{0}/databases/{1}/docs?id={2}".format(server_node.url, server_node.database,
                                                          Utils.quote_key(self.key))

    def set_response(self, response):
        try:
            response = response.json()
            if "Error" in response:
                raise exceptions.ErrorResponseException(response["Error"])
            return response
        except ValueError:
            raise exceptions.ErrorResponseException(
                "Failed to put document in the database please check the connection to the server")


class BatchCommand(RavenCommand):
    def __init__(self, commands):
        super(BatchCommand, self).__init__(method="POST")
        self.commands_array = commands
        self.files = None

    def create_request(self, server_node):
        data = []
        for command in self.commands_array:
            if not hasattr(command, 'command'):
                raise ValueError("Not a valid command")
            if command.type == "AttachmentPUT":
                if not self.files:
                    self.files = {}
                self.files[command.name] = (
                    command.name, command.stream, command.content_type, {"Command-Type": "AttachmentStream"})
            data.append(command.to_json())

        if self.files and len(self.files) > 0:
            self.use_stream = True

        self.url = "{0}/databases/{1}/bulk_docs".format(server_node.url, server_node.database)
        self.data = {"Commands": data}

    def set_response(self, response):
        if response is None:
            raise exceptions.InvalidOperationException(
                "Got null response from the server after doing a batch, "
                "something is very wrong. Probably a garbled response.")

        try:
            response = response.json()
            if "Error" in response:
                raise ValueError(response["Error"])
            return response["Results"]
        except ValueError as e:
            raise exceptions.InvalidOperationException(e)


class DeleteIndexCommand(RavenCommand):
    def __init__(self, index_name, database_name=None):
        """
        @param str index_name: Name of the index you like to get or delete
        @param str database_name: Name of the index database
        """
        super(DeleteIndexCommand, self).__init__(method="DELETE")
        self.index_name = index_name
        self.database_name = database_name

    def create_request(self, server_node):
        if not self.index_name:
            raise ValueError("None or empty index_name is invalid")

        database = self.database_name if self.database_name else server_node.database
        self.url = "{0}/databases/{1}/indexes?name={2}".format(server_node.url, database,
                                                               Utils.quote_key(self.index_name, True))

    def set_response(self, response):
        pass


class PatchCommand(RavenCommand):
    def __init__(self, document_id, change_vector, patch, patch_if_missing,
                 skip_patch_if_change_vector_mismatch=False, return_debug_information=False, test=False):
        """
        @param str document_id: The id of the document
        @param str change_vector: The change_vector
        @param PatchRequest patch: The patch that going to be applied on the document
        @param PatchRequest patch_if_missing: The default patch to applied
        @param bool skip_patch_if_change_vector_mismatch: If True will skip documents that mismatch the change_vector
        """
        if document_id is None:
            raise ValueError("None key is invalid")
        if patch is None:
            raise ValueError("None patch is invalid")
        if patch_if_missing and not patch_if_missing.script:
            raise ValueError("None or Empty script is invalid")

        super(PatchCommand, self).__init__(method="PATCH")
        self._document_id = document_id
        self._change_vector = change_vector
        self._patch = patch
        self._patch_if_missing = patch_if_missing
        self._skip_patch_if_change_vector_mismatch = skip_patch_if_change_vector_mismatch
        self._return_debug_information = return_debug_information
        self._test = test

    def create_request(self, server_node):
        path = "docs?id={0}".format(Utils.quote_key(self._document_id))
        if self._skip_patch_if_change_vector_mismatch:
            path += "&skipPatchIfChangeVectorMismatch=true"
        if self._return_debug_information:
            path += "&debug=true"
        if self._test:
            path += "&test=true"

        if self._change_vector is not None:
            self.headers = {"If-Match": "\"{0}\"".format(self._change_vector)}

        self.url = "{0}/databases/{1}/{2}".format(server_node.url, server_node.database, path)
        self.data = {"Patch": self._patch.to_json(),
                     "PatchIfMissing": self._patch_if_missing.to_json() if self._patch_if_missing else None}

    def set_response(self, response):
        if response is None:
            return None

        if response and response.status_code == 200:
            return response.json()


class QueryCommand(RavenCommand):
    def __init__(self, conventions, index_query, metadata_only=False, index_entries_only=False):
        """
        @param IndexQuery index_query: A query definition containing all information required to query a specified index.
        @param bool metadata_only: True if returned documents should include only metadata without a document body.
        @param bool index_entries_only: True if query results should contain only index entries.
        @return:json
        :rtype:dict
        """
        super(QueryCommand, self).__init__(method="POST", is_read_request=True)
        if index_query is None:
            raise ValueError("Invalid index_query")
        if conventions is None:
            raise ValueError("Invalid convention")
        self._conventions = conventions
        self._index_query = index_query
        self._metadata_only = metadata_only
        self._index_entries_only = index_entries_only

    def create_request(self, server_node):
        self.url = "{0}/databases/{1}/queries?query-hash={2}".format(server_node.url, server_node.database,
                                                                     self._index_query.get_query_hash())
        if self._metadata_only:
            self.url += "&metadataOnly=true"
        if self._index_entries_only:
            self.url += "&debug=entries"

        self.data = self._index_query.to_json()

    def set_response(self, response):
        if response is None:
            return

        response = response.json()
        if "Error" in response:
            raise exceptions.ErrorResponseException(response["Error"])
        return response


class GetStatisticsCommand(RavenCommand):
    def __init__(self, debug_tag=None):
        super(GetStatisticsCommand, self).__init__(method="GET")
        self.debug_tag = debug_tag

    def create_request(self, server_node):
        self.url = "{0}/databases/{1}/stats".format(server_node.url, server_node.database)
        if self.debug_tag:
            self.url += "?" + self.debug_tag

    def set_response(self, response):
        if response and response.status_code == 200:
            return response.json()
        return None


class GetTopologyCommand(RavenCommand):
    def __init__(self, force_url=None):
        super(GetTopologyCommand, self).__init__(method="GET", is_read_request=True)
        self._force_url = force_url

    def create_request(self, server_node):
        self.url = "{0}/topology?name={1}".format(server_node.url, server_node.database)
        if self._force_url is not None and not self._force_url == "":
            self.url += "&url={0}".format(self._force_url)

    def set_response(self, response):
        if response.status_code == 200:
            return response.json()
        if response.status_code == 400:
            log.debug(response.json()["Error"])
        return None


class GetClusterTopologyCommand(RavenCommand):
    def __init__(self):
        super(GetClusterTopologyCommand, self).__init__(method="GET", is_read_request=True)

    def create_request(self, server_node):
        self.url = "{0}/cluster/topology".format(server_node.url)

    def set_response(self, response):
        if response.status_code == 200:
            return response.json()
        if response.status_code == 400:
            log.debug(response.json()["Error"])
        return None


class GetOperationStateCommand(RavenCommand):
    def __init__(self, operation_id, is_server_store_operation=False):
        super(GetOperationStateCommand, self).__init__(method="GET")
        self.operation_id = operation_id
        self.is_server_store_operation = is_server_store_operation

    def create_request(self, server_node):
        self.url = "{0}/databases/{1}/operations/state?id={2}".format(server_node.url, server_node.database,
                                                                      self.operation_id) if not self.is_server_store_operation \
            else "{0}/operations/state?id={2}".format(server_node.url, self.operation_id)

    def set_response(self, response):
        try:
            response = response.json()
        except ValueError:
            raise response.raise_for_status()
        return response


class PutAttachmentCommand(RavenCommand):
    def __init__(self, document_id, name, stream, content_type, change_vector):
        """
        @param document_id: The id of the document
        @param name: Name of the attachment
        @param stream: The attachment as bytes (ex.open("file_path", "rb"))
        @param content_type: The type of the attachment (ex.image/png)
        @param change_vector: The change vector of the document
        """

        if not document_id:
            raise ValueError("Invalid document_id")
        if not name:
            raise ValueError("Invalid name")

        super(PutAttachmentCommand, self).__init__(method="PUT")
        self._document_id = document_id
        self._name = name
        self._stream = stream
        self._content_type = content_type
        self._change_vector = change_vector

    def create_request(self, server_node):
        self.url = "{0}/databases/{1}/attachments?id={2}&name={3}".format(server_node.url, server_node.database,
                                                                          Utils.quote_key(self._document_id),
                                                                          Utils.quote_key(self._name))
        if self._content_type:
            self.url += "&contentType={0}".format(Utils.quote_key(self._content_type))
        self.data = self._stream

        if self._change_vector is not None:
            self.headers = {"If-Match": "\"{0}\"".format(self._change_vector)}

    def set_response(self, response):
        try:
            response = response.json()
            if "Error" in response:
                raise exceptions.InvalidOperationException(response["Error"])
            return response
        except ValueError:
            raise response.raise_for_status()


class GetFacetsCommand(RavenCommand):
    def __init__(self, query):
        """
        @param FacetQuery query: The query we wish to get
        """
        if not query:
            raise ValueError("Invalid query")
        super(GetFacetsCommand, self).__init__(method="POST", is_read_request=True)
        self._query = query
        if query.wait_for_non_stale_results_timeout and query.wait_for_non_stale_results_timeout != timedelta.max:
            self.timeout = self._query.wait_for_non_stale_results_timeout + timedelta(seconds=10)

    def create_request(self, server_node):
        if self._query.facet_setup_doc and len(self._query.facets) > 0:
            raise exceptions.InvalidOperationException("You cannot specify both 'facet_setup_doc' and 'facets'.")

        self.url = "{0}/databases/{1}/queries?op=facets&query-hash={2}".format(server_node.url, server_node.database,
                                                                               self._query.get_query_hash())
        self.data = self._query.to_json()

    def set_response(self, response):
        if response.status_code == 200:
            return response.json()


class MultiGetCommand(RavenCommand):
    def __init__(self, requests):
        """
        @param requests: The requests for the server
        :type list of dict
        dict keys:
            url = Request url (relative).
            headers =  Request headers.
            query = Query information e.g. "?pageStart=10&amp;pageSize=20".
            data = The data for the requests body
            method = The requests method e.g.
        """
        super(MultiGetCommand, self).__init__(method="POST")
        self._requests = requests
        self._base_url = None

    def create_request(self, server_node):
        self._base_url = "{0}/databases/{1}".format(server_node.url, server_node.database)
        commands = []
        for request in self._requests:
            headers = {}
            for key, value in request.get("headers", {}).items():
                headers[key] = value
            commands.append(
                {"Url": "/databases/{0}{1}".format(server_node.database, request["url"]),
                 "Query": request.get("query", None), "Method": request.get("method", None),
                 "Content": request.get("data", None), "Headers": headers})

        self.data = {"Requests": commands}
        self.url = "{0}/multi_get".format(self._base_url)

    def set_response(self, response):
        try:
            response = response.json()
            if "Error" in response:
                raise exceptions.ErrorResponseException(response["Error"])
            return response["Results"]
        except:
            raise exceptions.ErrorResponseException("Invalid response")


class GetDatabaseRecordCommand(RavenCommand):
    def __init__(self, database_name):
        super(GetDatabaseRecordCommand, self).__init__(method="GET")
        self._database_name = database_name

    def create_request(self, server_node):
        self.url = "{0}/admin/databases?name={1}".format(server_node.url, self._database_name)

    def set_response(self, response):
        try:
            response = response.json()
            if "Error" in response:
                raise exceptions.ErrorResponseException(response["Error"])
            return response["Topology"]
        except:
            raise response.raise_for_status()


class WaitForRaftIndexCommand(RavenCommand):
    def __init__(self, index):
        super(WaitForRaftIndexCommand, self).__init__(method="GET", is_read_request=True)
        self._index = index

    def create_request(self, server_node):
        self.url = "{0}/rachis/waitfor?index={1}".format(server_node.url, self._index)

    def set_response(self, response):
        if response is None:
            raise exceptions.ErrorResponseException("Invalid response")


class GetTcpInfoCommand(RavenCommand):
    def __init__(self, tag, database_name=None):
        super(GetTcpInfoCommand, self).__init__(method="GET")
        self._tag = tag
        self._database_name = database_name

    def create_request(self, server_node):
        if self._database_name is None:
            self.url = "{0}/info/tcp?tag={1}".format(server_node.url, self._tag)
        else:
            self.url = "{0}/databases/{1}/info/tcp?tag={2}".format(server_node.url, self._database_name, self._tag)

        self.requested_node = server_node

    def set_response(self, response):
        if response is None:
            raise exceptions.ErrorResponseException("Invalid response")

        return response.json()


class QueryStreamCommand(RavenCommand):
    def __init__(self, index_query):
        super(QueryStreamCommand, self).__init__(method="POST", is_read_request=True, use_stream=True)
        if not index_query:
            raise ValueError("index_query cannot be None")

        if index_query.wait_for_non_stale_results:
            raise exceptions.NotSupportedException("Since stream() does not wait for indexing (by design), "
                                                   "streaming query with wait_for_non_stale_results is not supported.")

        self._index_query = index_query

    def create_request(self, server_node):
        self.url = "{0}/databases/{1}/streams/queries".format(server_node.url, server_node.database)
        self.data = self._index_query.to_json()

    def set_response(self, response):
        if response is None:
            raise exceptions.ErrorResponseException("Invalid response")
        if response.status_code != 200:
            try:
                json_response = response.json()
                if "Error" in json_response:
                    raise Exception(json_response["Error"])
                return response
            except ValueError:
                raise response.raise_for_status()
        return response


# ------------------------SubscriptionCommands----------------------

class CreateSubscriptionCommand(RavenCommand):
    def __init__(self, options):
        """
        @param SubscriptionCreationOptions options: Subscription options
        """

        super(CreateSubscriptionCommand, self).__init__(method="PUT")
        self._options = options

    def create_request(self, server_node):
        self.url = "{0}/databases/{1}/subscriptions".format(server_node.url, server_node.database)
        self.data = self._options.to_json()

    def set_response(self, response):
        try:
            response = response.json()
            if "Error" in response:
                raise exceptions.ErrorResponseException(response["Error"])
            return response["Name"]
        except:
            raise response.raise_for_status()


class DeleteSubscriptionCommand(RavenCommand):
    def __init__(self, name):
        super(DeleteSubscriptionCommand, self).__init__(method="DELETE")
        self._name = name

    def create_request(self, server_node):
        self.url = "{0}/databases/{1}/subscriptions?taskName={2}".format(server_node.url, server_node.database,
                                                                         self._name)

    def set_response(self, response):
        pass


class DropSubscriptionConnectionCommand(RavenCommand):
    def __init__(self, name):
        super(DropSubscriptionConnectionCommand, self).__init__(method="POST")
        self._name = name

    def create_request(self, server_node):
        self.url = "{0}/databases/{1}/subscriptions/drop?name={2}".format(server_node.url, server_node.database,
                                                                          self._name)

    def set_response(self, response):
        pass


class GetSubscriptionsCommand(RavenCommand):
    def __init__(self, start, page_size):
        super(GetSubscriptionsCommand, self).__init__(method="GET", is_read_request=True)
        self._start = start
        self._page_size = page_size

    def create_request(self, server_node):
        self.url = "{0}/databases/{1}/subscriptions?start={2}&pageSize={3}".format(
            server_node.url, server_node.database, self._start, self._page_size)

    def set_response(self, response):
        if response is None:
            raise ValueError("response is invalid.")
        inner_data = {}
        data = []
        try:
            response = response.json()
            if "Error" in response:
                raise exceptions.ErrorResponseException(response["Error"])
            for item in response["Results"]:
                for key, value in item.items():
                    inner_data[Utils.convert_to_snake_case(key)] = value
                data.append(SubscriptionState(**inner_data))
            return data

        except ValueError:
            raise response.raise_for_status()


class GetSubscriptionStateCommand(RavenCommand):
    def __init__(self, subscription_name):
        super(GetSubscriptionStateCommand, self).__init__(method="GET", is_read_request=True)
        self._subscription_name = subscription_name

    def create_request(self, server_node):
        self.url = "{0}/databases/{1}/subscriptions/state?name={2}".format(
            server_node.url, server_node.database, self._subscription_name)

    def set_response(self, response):
        if response is None:
            raise ValueError("response is invalid.")
        data = {}
        try:
            response = response.json()
            if "Error" in response:
                raise exceptions.ErrorResponseException(response["Error"])

            for key, value in response.items():
                data[Utils.convert_to_snake_case(key)] = value
            return SubscriptionState(**data)

        except ValueError:
            raise response.raise_for_status()



from pyravendb.custom_exceptions import exceptions
import OpenSSL.crypto
from collections import Iterable
from datetime import datetime, timedelta
from threading import Timer
import urllib
import inspect
import json
import sys
import re


class _DynamicStructure(object):
    def __init__(self, **entries):
        self.__dict__.update(entries)

    def __str__(self):
        return str(self.__dict__)


class Utils(object):
    @staticmethod
    def quote_key(key, reserved_slash=False):
        reserved = '%:=&?~#+!$,;\'*[]'
        if reserved_slash:
            reserved += '/'
        if key:
            return urllib.parse.quote(key, safe=reserved)
        else:
            return ''

    @staticmethod
    def unpack_iterable(iterable):
        for item in iterable:
            if isinstance(item, Iterable) and not isinstance(item, str):
                for nested in Utils.unpack_iterable(item):
                    yield nested
            else:
                yield item

    @staticmethod
    def convert_to_snake_case(name):
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

    @staticmethod
    def database_name_validation(name):
        if name is None:
            raise ValueError("None name is not valid")
        result = re.match(r'([A-Za-z0-9_\-\.]+)', name, re.IGNORECASE)
        if not result:
            raise exceptions.InvalidOperationException(
                "Database name can only contain only A-Z, a-z, \"_\", \".\" or \"-\" but was: " + name)

    @staticmethod
    def first_or_default(iterator, func, default):
        for item in iterator:
            if func(item):
                return item
        return default

    @staticmethod
    def get_change_vector_from_header(response):
        header = response.headers.get("ETag", None)
        if header is not None and header[0] == "\"":
            return header[1: len(header) - 2]

    @staticmethod
    def import_class(name):
        components = name.split('.')
        module_name = '.'.join(name.split('.')[:-1])
        mod = None
        try:
            mod = getattr(__import__(module_name, fromlist=[components[-1]]), components[-1])
        except (ImportError, ValueError, AttributeError):
            pass
        return mod

    @staticmethod
    def is_inherit(parent, child):
        if parent == child:
            return True
        if child is None:
            return False
        if parent != child:
            return Utils.is_inherit(parent, child.__base__)

    @staticmethod
    def convert_to_entity(document, object_type, conventions, nested_object_types=None, fetch=False):
        metadata = document.pop("@metadata")
        original_metadata = metadata.copy()
        type_from_metadata = conventions.try_get_type_from_metadata(metadata)
        if object_type == dict:
            return document, metadata, original_metadata
        entity = _DynamicStructure(**document)
        if not fetch:
            object_from_metadata = None
            if type_from_metadata is not None:
                object_from_metadata = Utils.import_class(type_from_metadata)

            if object_from_metadata is None:
                if object_type is not None:
                    entity.__class__ = object_type
                    metadata["Raven-Python-Type"] = "{0}.{1}".format(object_type.__module__, object_type.__name__)
            else:
                if object_type and not Utils.is_inherit(object_type, object_from_metadata):
                    raise exceptions.InvalidOperationException(
                        "Unable to cast object of type {0} to type {1}".format(object_from_metadata, object_type))
                entity.__class__ = object_from_metadata
            # Checking the class for initialize
            entity_initialize_dict = Utils.make_initialize_dict(document, entity.__class__.__init__)

            temp_entity = entity.__class__(**entity_initialize_dict)
            for key, value in entity.__dict__.items():
                if key not in entity_initialize_dict:
                    if hasattr(temp_entity, key):
                        setattr(temp_entity, key, value)
            entity = temp_entity

        if nested_object_types:
            for key in nested_object_types:
                attr = getattr(entity, key)
                if attr:
                    try:
                        if isinstance(attr, list):
                            nested_list = []
                            for attribute in attr:
                                nested_list.append(nested_object_types[key](
                                    **Utils.make_initialize_dict(attribute, nested_object_types[key].__init__)))
                            setattr(entity, key, nested_list)

                        elif nested_object_types[key] is datetime:
                            setattr(entity, key, Utils.string_to_datetime(attr))
                        elif nested_object_types[key] is timedelta:
                            setattr(entity, key, Utils.string_to_timedelta(attr))
                        else:
                            setattr(entity, key, nested_object_types[key](
                                **Utils.make_initialize_dict(attr, nested_object_types[key].__init__)))
                    except TypeError:
                        pass

        if 'Id' in entity.__dict__:
            entity.Id = metadata.get('@id', None)
        return entity, metadata, original_metadata

    @staticmethod
    def make_initialize_dict(document, entity_init):
        if entity_init is None:
            return document

        entity_initialize_dict = {}
        args, __, keywords, defaults, _, _, _ = inspect.getfullargspec(entity_init)
        if (len(args) - 1) > len(document):
            remainder = len(args)
            if defaults:
                remainder -= len(defaults)
            for i in range(1, remainder):
                entity_initialize_dict[args[i]] = document.get(args[i], None)
            for i in range(remainder, len(args)):
                entity_initialize_dict[args[i]] = document.get(args[i], defaults[i - remainder])
        else:
            if keywords:
                entity_initialize_dict = document
            else:
                for key in document:
                    if key in args:
                        entity_initialize_dict[key] = document[key]

        return entity_initialize_dict

    @staticmethod
    def dict_to_bytes(the_dict):
        json_dict = json.dumps(the_dict)
        return bytes(json_dict, encoding='utf-8')

    @staticmethod
    def dict_to_string(dictionary):
        builder = []
        for item in dictionary:
            if sys.version_info.major > 2 and isinstance(dictionary[item], bytes):
                dictionary[item] = dictionary[item].decode('utf-8')
            builder.append('{0}={1}'.format(item, dictionary[item]))
        return ','.join(item for item in builder)

    @staticmethod
    def datetime_to_string(datetime_obj):
        return datetime_obj.strftime("%Y-%m-%dT%H:%M:%S.%f0")

    @staticmethod
    def start_a_timer(interval, function, args=None, name=None, daemon=False):
        timer = Timer(interval, function, args)
        timer.daemon = daemon
        if name is not None:
            timer.name = name
        timer.start()

        return timer

    @staticmethod
    def string_to_datetime(datetime_str):
        try:
            if datetime_str.endswith('Z'):
                datetime_str = datetime_str[:-1]
            datetime_s = datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S.%f")
        except ValueError:
            datetime_s = datetime.strptime(datetime_str[:-1], "%Y-%m-%dT%H:%M:%S.%f")

        return datetime_s

    @staticmethod
    def timedelta_tick(td):
        return int(td.total_seconds() * 10000000)

    @staticmethod
    def string_to_timedelta(timedelta_str):
        pattern = r'(?:(-?\d+)[.])?(\d{2}):(\d{2}):(\d{2})(?:.(\d+))?'
        timedelta_initialize = None
        m = re.match(pattern, timedelta_str, re.IGNORECASE)
        if m:
            timedelta_initialize = {"days": 0 if m.group(1) is None else int(m.group(1)),
                                    "hours": 0 if m.group(2) is None else int(m.group(2)),
                                    "minutes": 0 if m.group(3) is None else int(m.group(3)),
                                    "seconds": 0 if m.group(4) is None else int(m.group(4)),
                                    "microseconds": 0 if m.group(5) is None else int(m.group(5))
                                    }
        if timedelta_initialize:
            return timedelta(**timedelta_initialize)
        return None

    @staticmethod
    def timedelta_to_str(timedelta_obj):
        timedelta_str = ""
        if isinstance(timedelta_obj, timedelta):
            total_seconds = timedelta_obj.seconds
            days = timedelta_obj.days
            hours = total_seconds // 3600
            minutes = (total_seconds // 60) % 60
            seconds = (total_seconds % 3600) % 60
            microseconds = timedelta_obj.microseconds
            if days > 0:
                timedelta_str += "{0}.".format(days)
            timedelta_str += "{:02}:{:02}:{:02}".format(hours, minutes, seconds)
            if microseconds > 0:
                timedelta_str += ".{0}".format(microseconds)
        return timedelta_str

    @staticmethod
    def escape(term, allow_wild_cards, make_phrase):
        wild_cards = ['-', '&', '|', '!', '(', ')', '{', '}', '[', ']', '^', '"', '~', ':', '\\']
        if not term:
            return "\"\""
        start = 0
        length = len(term)
        buffer = ""
        if length >= 2 and term[0] == '/' and term[1] == '/':
            buffer += "//"
            start = 2
        i = start
        while i < length:
            ch = term[i]
            if ch == '*' or ch == '?':
                if allow_wild_cards:
                    i += 1
                    continue

            if ch in wild_cards:
                if i > start:
                    buffer += term[start:i - start]

                buffer += '\\{0}'.format(ch)
                start = i + 1

            elif ch == ' ' or ch == '\t':
                if make_phrase:
                    return "\"{0}\"".format(Utils.escape(term, allow_wild_cards, False))

            i += 1
        if length > start:
            buffer += term[start: length - start]

        return buffer

    @staticmethod
    def pfx_to_pem(pem_path, pfx_path, pfx_password):
        """
        @param pem_path: Where to create the pem file
        @param pfx_path: The path to the pfx file
        @param pfx_password: The password to pfx file
        """
        with open(pem_path, 'wb') as pem_file:
            with open(pfx_path, 'rb') as pfx_file:
                pfx = pfx_file.read()
            p12 = OpenSSL.crypto.load_pkcs12(pfx, pfx_password)
            pem_file.write(OpenSSL.crypto.dump_privatekey(OpenSSL.crypto.FILETYPE_PEM, p12.get_privatekey()))
            pem_file.write(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, p12.get_certificate()))
            ca = p12.get_ca_certificates()
            if ca is not None:
                for cert in ca:
                    # In python 3.6* we need to save the ca to ?\lib\site-packages\certifi\cacert.pem.
                    pem_file.write(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, cert))
        return pem_path

    @staticmethod
    def get_cert_file_fingerprint(pem_path):
        with open(pem_path, 'rb') as pem_file:
            cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, pem_file.read())
            return str(cert.digest("sha1"))

    @staticmethod
    def index_of_any(text, any_of):
        """
        @param str text: The text we want to check
        @param list any_of: list of char
        :returns False if text nit
        """
        result = -1
        if not text or not any_of:
            return result

        any_of = list(set(any_of))
        i = 0
        while i < len(text) and result == -1:
            for c in any_of:
                if c == text[i]:
                    result = i
                    break
            i += 1
        return result

    @staticmethod
    def contains_any(text, any_of):
        """
        @param text: The text we want to check
        :type str
        @param any_of: list of char
        :type list
        :returns False if text nit
        """
        if not text or not any_of:
            return False

        any_of = list(set(any_of))
        for c in any_of:
            if c in text:
                return True
        return False



from threading import Lock


class ConcurrentSet(set):
    def __init__(self, seq=()):
        super(ConcurrentSet, self).__init__(seq)
        self._lock = Lock()

    def add(self, element):
        with self._lock:
            super(ConcurrentSet, self).add(element)

    def discard(self, element):
        with self._lock:
            super(ConcurrentSet, self).discard(element)



from pyravendb.tests.test_base import TestBase
from pyravendb.data.indexes import IndexDefinition
from pyravendb.raven_operations.maintenance_operations import PutIndexesOperation
from pyravendb.changes.observers import ActionObserver
import unittest
from threading import Event
from time import sleep


class User:
    def __init__(self, name):
        self.name = name


class Address:
    def __init__(self, country):
        self.country = country


class TestDatabaseChanges(TestBase):
    def setUp(self):
        super(TestDatabaseChanges, self).setUp()

    def tearDown(self):
        super(TestDatabaseChanges, self).tearDown()
        self.delete_all_topology_files()

    def test_for_all_documents(self):
        event = Event()
        all_documents_changes = []

        def on_next(value):
            all_documents_changes.append(value)
            if len(all_documents_changes) == 9:
                event.set()

        all_observer = self.store.changes().for_all_documents()
        all_observer.subscribe(ActionObserver(on_next=on_next))
        all_observer.ensure_subscribe_now()
        with self.store.open_session() as session:
            session.store(User("Idan"))
            session.store(User("Shalom"))
            session.store(User("Ilay"))
            session.store(User("Agam"))
            session.store(User("Lee"))
            session.store(User("Oren"))
            session.store(User("Ayende"))
            session.store(User("Raven"))
            session.save_changes()
        event.wait(10)
        self.assertEqual(9, len(all_documents_changes))

    def test_for_document(self):
        documents = []
        observer = self.store.changes().for_document("users/1")
        observer.subscribe(documents.append)
        observer.ensure_subscribe_now()

        with self.store.open_session() as session:
            session.store(User("Idan"), key="users/1")
            session.store(User("Shalom"), key="users/2")
            session.save_changes()

        sleep(1)
        self.assertTrue(len(documents) == 1)
        self.assertEqual("users/1", documents[0]["Id"])
        self.assertEqual('Put', documents[0]["Type"])
        self.assertIsNotNone(documents[0]["ChangeVector"])

    def test_document_change(self):
        documents = []
        observer = self.store.changes().for_document("users/1")
        observer.subscribe(documents.append)
        observer.ensure_subscribe_now()

        with self.store.open_session() as session:
            session.store(User("Idan"), key="users/1")
            session.save_changes()

        with self.store.open_session() as session:
            user = session.load("users/1")
            user.name = "Ilay"
            session.save_changes()

        sleep(1)
        self.assertTrue(len(documents) == 2)
        self.assertEqual(documents[1]["Id"], documents[0]["Id"])
        self.assertNotEqual(documents[1]["ChangeVector"], documents[0]["ChangeVector"])

    def test_for_all_indexes(self):
        indexes_change = []
        observer = self.store.changes().for_all_indexes()
        observer.subscribe(indexes_change.append)
        observer.ensure_subscribe_now()

        index_definition_dogs = IndexDefinition("Dogs", "from doc in docs.Dogs select new {doc.brand}")

        self.store.maintenance.send(PutIndexesOperation(index_definition_dogs))

        sleep(1)
        self.assertTrue(len(indexes_change) >= 1)
        self.assertEqual(indexes_change[0]['Name'], 'Dogs')
        self.assertEqual(indexes_change[0]['Type'], 'IndexAdded')

    def test_for_index(self):
        index_change = []
        observer = self.store.changes().for_index('Dogs')
        observer.subscribe(index_change.append)
        observer.ensure_subscribe_now()

        index_definition_users = IndexDefinition("Users", "from doc in docs.Users select new {doc.Name}")
        index_definition_dogs = IndexDefinition("Dogs", "from doc in docs.Dogs select new {doc.brand}")

        self.store.maintenance.send(PutIndexesOperation(index_definition_dogs, index_definition_users))

        sleep(1)
        for index in index_change:
            self.assertFalse(index['Name'] != 'Dogs')

    def test_subscribe_to_index_and_document(self):
        documents = []
        indexes = []

        observer = self.store.changes().for_document("users/1")
        observer.subscribe(documents.append)
        observer.ensure_subscribe_now()

        observer = self.store.changes().for_index('Dogs')
        observer.subscribe(indexes.append)
        observer.ensure_subscribe_now()

        index_definition_users = IndexDefinition("Users", "from doc in docs.Users select new {doc.Name}")
        index_definition_dogs = IndexDefinition("Dogs", "from doc in docs.Dogs select new {doc.brand}")

        self.store.maintenance.send(PutIndexesOperation(index_definition_dogs, index_definition_users))

        with self.store.open_session() as session:
            session.store(User("Idan"), key="users/1")
            session.save_changes()

        sleep(1)
        self.assertTrue(len(documents) >= 1)
        self.assertTrue(len(indexes) >= 1)

    def test_for_documents_start_with(self):
        documents = []
        observer = self.store.changes().for_documents_start_with("users")
        observer.subscribe(documents.append)
        observer.ensure_subscribe_now()

        with self.store.open_session() as session:
            session.store(User("Idan"))
            session.store(User("Shalom"))
            session.store(User("Ayende"))
            session.store(Address("Israel"))
            session.save_changes()

        sleep(1)
        observer.ensure_subscribe_now()
        self.assertEqual(3, len(documents))
        for document in documents:
            self.assertTrue(document['Id'].startswith('users'))


if __name__ == "__main__":
    unittest.main()



from pyravendb.custom_exceptions.exceptions import InvalidOperationException
from pyravendb.commands.raven_commands import CreateSubscriptionCommand, DeleteSubscriptionCommand, \
    GetSubscriptionsCommand, GetSubscriptionStateCommand, DropSubscriptionConnectionCommand
from threading import Lock
from pyravendb.subscriptions.subscription import SubscriptionWorker
from pyravendb.subscriptions.data import *


class DocumentSubscriptions:
    def __init__(self, store):
        self._store = store
        self.lock = Lock()
        self._subscriptions = set()

    def create(self, options_or_query, database=None):
        """
        @param str or SubscriptionCreationOptions options_or_query: The query or the options to create the subscription
        @param str database: The database name
        :returns: returns The subscription name
        """
        options = options_or_query
        if isinstance(options_or_query, str):
            options = SubscriptionCreationOptions(query=options_or_query)

        return self._create(options, database)

    def _create(self, options, database=None):
        if options is None:
            raise InvalidOperationException("Cannot create a subscription if the options is set to null")
        if not isinstance(options, SubscriptionCreationOptions):
            raise InvalidOperationException("options must be SubscriptionCreationOptions")
        if options.query is None:
            raise InvalidOperationException("Cannot create a subscription if the script is null")

        request_executor = self._store.get_request_executor(database)
        command = CreateSubscriptionCommand(options)
        return request_executor.execute(command)

    def get_subscription_worker(self, options, database=None, object_type=None, nested_object_types=None):
        """
        @param SubscriptionWorkerOptions options: The subscription connection options.
        @param str database: The database name
        @param Type object_type: The type of the object we want to track the entity to
        @param Dict[str, Type] nested_object_types: A dict of classes for nested object the key will be the name of the
        class and the value will be the object we want to get for that attribute
        """
        if options is None:
            raise InvalidOperationException("Cannot create a subscription if the options is set to None")
        if not isinstance(options, SubscriptionWorkerOptions):
            raise ValueError("Invalid options", "options mus be SubscriptionWorkerOptions")

        subscription = SubscriptionWorker(options, self._store, database, object_type=object_type,
                                          nested_object_types=nested_object_types)
        with self.lock:
            self._subscriptions.add(subscription)

        return subscription

    def delete(self, name_or_id, database=None):
        request_executor = self._store.get_request_executor(database)
        command = DeleteSubscriptionCommand(name_or_id)
        request_executor.execute(command)

    def drop_connection(self, name, database=None):
        """
        Force server to close current client subscription connection to the server
        @param str name: The name of the subscription
        @param str database: The name of the database
        """
        request_executor = self._store.get_request_executor(database)
        command = DropSubscriptionConnectionCommand(name)
        request_executor.execute(command)

    def get_subscriptions(self, start, take, database=None):
        """
        @param int start: Where to start and take the subscriptions
        @param int take: How much subscription to get
        @param str database: The database name
        @returns: returns a list of SubscriptionState
        """
        request_executor = self._store.get_request_executor(database)
        command = GetSubscriptionsCommand(start, take)
        return request_executor.execute(command)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if len(self._subscriptions) == 0:
            return

        for subscription in self._subscriptions:
            subscription.close()

    def get_subscription_state(self, subscription_name, database=None):
        request_executor = self._store.get_request_executor(database)
        command = GetSubscriptionStateCommand(subscription_name)
        return request_executor.execute(command)



import unittest
from pyravendb.data.indexes import IndexDefinition
from pyravendb.raven_operations.maintenance_operations import PutIndexesOperation, GetIndexOperation
from pyravendb.commands.raven_commands import DeleteIndexCommand
from pyravendb.tests.test_base import TestBase


class TestIndexActions(TestBase):
    def setUp(self):
        super(TestIndexActions, self).setUp()
        self.requests_executor = self.store.get_request_executor()
        self.conventions = self.store.conventions

    def tearDown(self):
        super(TestIndexActions, self).tearDown()
        self.delete_all_topology_files()

    def test_put_index_success(self):
        index = IndexDefinition(name="region", maps=self.index_map)
        assert self.requests_executor.execute(PutIndexesOperation(index).get_command(self.conventions))

    def test_get_index_success(self):
        index = IndexDefinition(name="get_index", maps=self.index_map)
        assert self.requests_executor.execute(PutIndexesOperation(index).get_command(self.conventions))
        self.assertIsNotNone(
            self.requests_executor.execute(GetIndexOperation("get_index").get_command(self.conventions)))

    def test_get_index_fail(self):
        self.assertIsNone(self.requests_executor.execute(GetIndexOperation("get_index").get_command(self.conventions)))

    def test_delete_index_success(self):
        index = IndexDefinition(name="delete", maps=self.index_map)
        assert self.requests_executor.execute(PutIndexesOperation(index).get_command(self.conventions))
        self.assertIsNone(self.requests_executor.execute(DeleteIndexCommand("delete")))

    def test_delete_index_fail(self):
        with self.assertRaises(ValueError):
            self.requests_executor.execute(DeleteIndexCommand(None))


if __name__ == "__main__":
    unittest.main()



from pyravendb.tests.test_base import TestBase
from pyravendb.raven_operations.maintenance_operations import PutIndexesOperation
from pyravendb.custom_exceptions import exceptions
from pyravendb.data.indexes import IndexDefinition, SortOptions, IndexFieldOptions
import unittest


class User:
    def __init__(self, name, age):
        self.name = name
        self.age = age


class Product(object):
    def __init__(self, name, key, order):
        self.name = name
        self.key = key
        self.order = order


class Company(object):
    def __init__(self, name, product):
        self.name = name
        self.product = product


class Order(object):
    def __init__(self, name="", key=None, product_id=None):
        self.name = name
        self.key = key
        self.product_id = product_id


class TestQuery(TestBase):
    def setUp(self):
        super(TestQuery, self).setUp()
        index_map = ("from doc in docs "
                     "select new {"
                     "name = doc.name,"
                     "key = doc.key,"
                     "doc_id = doc.key_doc.name}")
        index_definition = IndexDefinition(name="Testing_Sort", maps=index_map,
                                           fields={"key": IndexFieldOptions(sort_options=SortOptions.numeric),
                                                   "doc_id": IndexFieldOptions(storage=True)})

        maps = ("from order in docs.Orders "
                "select new {"
                "key = order.key,"
                "order_and_id = order.name+\"_\"+order.product_id}")
        second_index_definition = IndexDefinition(name="SelectTestingIndex", maps=maps,
                                                  fields={"order_and_id": IndexFieldOptions(storage=True)})

        self.store.maintenance.send(PutIndexesOperation(index_definition, second_index_definition))

        with self.store.open_session() as session:
            session.store(Product("test101", 2, "a"), "products/101")
            session.store(Product("test10", 3, "b"), "products/10")
            session.store(Product("test106", 4, "c"), "products/106")
            session.store(Product("test107", 5, None), "products/107")
            session.store(Product("test107", 6, None), "products/103")
            session.store(Product("new_testing", 90, "d"), "products/108")
            session.store(Order("testing_order", 92, "products/108"), "orders/105")
            session.store(Company("withNesting", Product(name="testing_order", key=4, order=None)), "company/1")
            session.save_changes()

    def tearDown(self):
        super(TestQuery, self).tearDown()
        self.delete_all_topology_files()

    def test_where_equal_dynamic_index(self):
        with self.store.open_session() as session:
            query_results = list(session.query(collection_name="Products").where_equals("name", "test101"))
            self.assertEqual(query_results[0].name, "test101")

    def test_where_not_equal(self):
        with self.store.open_session() as session:
            query_results = list(session.query(collection_name="Products").where_not_equals("name", "test101"))
            for result in query_results:
                assert result.name != "test101"

    def test_where_equal_double(self):
        with self.store.open_session() as session:
            query_results = list(session.query().where_equals("name", "test101").where_equals("key", 4))
            self.assertEqual(len(query_results), 2)

    def test_where_equal_double_and_operator(self):
        with self.store.open_session() as session:
            query_results = list(
                session.query(object_type=Product).where_equals(
                    "name", "test107").and_also().where_equals("key", 5))
            self.assertEqual(len(query_results), 1)

    def test_where_in(self):
        with self.store.open_session() as session:
            names = ["test101", "test107", "test106"]
            query_result = list(session.query().where_in("name", names))
            self.assertEqual(len(query_result), 4)
            for entity in query_result:
                assert entity.name in names

    def test_where_starts_with(self):
        with self.store.open_session() as session:
            query_result = list(session.query(Product).where_starts_with("name", "n"))
            self.assertEqual(query_result[0].name, "new_testing")

    def test_where_ends_with(self):
        with self.store.open_session() as session:
            query_result = list(session.query(Product).where_ends_with("name", "7"))
            self.assertEqual(query_result[0].name, "test107")

    def test_query_return_empty_query_with_not_exist_index(self):
        with self.store.open_session() as session:
            self.assertEqual(len(list(session.query(index_name="s").where(Tag="Products"))), 0)

    def test_query_success_with_where(self):
        with self.store.open_session() as session:
            query_results = list(session.query().where(name="test101", key=[4, 6, 90]))
            self.assertGreater(len(query_results), 0)

    def test_query_success_with_index(self):
        with self.store.open_session() as session:
            keys = [4, 6, 90]
            query_results = list(
                session.query(index_name="Testing_Sort", wait_for_non_stale_results=True).where(key=keys))
            self.assertEqual(len(query_results), 3)
            for result in query_results:
                assert result.key in keys

    def test_where_between(self):
        with self.store.open_session() as session:
            query_results = list(
                session.query(index_name="Testing_Sort", wait_for_non_stale_results=True).where_between("key", 2, 4))
            for result in query_results:
                self.assertGreaterEqual(result.key, 2)
                self.assertLessEqual(result.key, 4)

    def test_query_with_order_by(self):
        with self.store.open_session() as session:
            query_results = list(
                session.query(wait_for_non_stale_results=True).where_not_none("order").order_by("order"))
            self.assertEqual(query_results[0].order, "a")

    def test_query_with_order_by_descending(self):
        with self.store.open_session() as session:
            query_results = list(
                session.query(wait_for_non_stale_results=True).where_not_none("order").order_by_descending("order"))
            self.assertEqual(query_results[0].order, "d")

    def test_where_not_None(self):
        found_none = False
        with self.store.open_session() as session:
            query_results = list(session.query(wait_for_non_stale_results=True).where_not_none("order"))
            for result in query_results:
                if result.order is None:
                    found_none = True
                    break
            self.assertFalse(found_none)

    def test_where_with_include(self):
        with self.store.open_session() as session:
            list(session.query(wait_for_non_stale_results=True).where(key=92).include("product_id"))
            session.load("products/108")
        self.assertEqual(session.number_of_requests_in_session, 1)

    def test_query_with_nested_object(self):
        with self.store.open_session() as session:
            query_results = list(
                session.query(object_type=Company, nested_object_types={"product": Product}).where_equals(
                    "name", "withNesting"))
            self.assertTrue(isinstance(query_results[0], Company) and isinstance(query_results[0].product, Product))

    def test_query_with_raw_query(self):
        with self.store.open_session() as session:
            query_results = list(
                session.query(object_type=Company, nested_object_types={"product": Product}).raw_query(
                    "FROM Companies WHERE name='withNesting'"))
            self.assertTrue(isinstance(query_results[0], Company) and isinstance(query_results[0].product, Product))

    def test_raw_query_with_query_parameters(self):
        with self.store.open_session() as session:
            query_results = list(
                session.query(object_type=Company, nested_object_types={"product": Product}).raw_query(
                    "FROM Companies WHERE name= $p0", query_parameters={"p0": "withNesting"}))
            self.assertTrue(isinstance(query_results[0], Company) and isinstance(query_results[0].product, Product))

    def test_fail_after_raw_query_add(self):
        try:
            with self.store.open_session() as session:
                query_results = list(
                    session.query(object_type=Company, nested_object_types={"product": Product}).raw_query(
                        "FROM Companies WHERE name='withNesting'").where_starts_with("product", 'p'))
                self.assertTrue(isinstance(query_results[0], Company) and isinstance(query_results[0].product, Product))
        except exceptions.InvalidOperationException as e:
            self.assertTrue("raw_query was called" in str(e))

    def test_query_with_select(self):
        with self.store.open_session() as session:
            query_results = list(
                session.query(object_type=dict, wait_for_non_stale_results=True, index_name="SelectTestingIndex").where(
                    key=92).select("order_and_id"))
            for result in query_results:
                self.assertEqual(len(result), 1)
                self.assertTrue("order_and_id" in result)


if __name__ == "__main__":
    unittest.main()



import unittest
from pyravendb.data.document_conventions import DocumentConventions
from pyravendb.commands.raven_commands import *
from pyravendb.custom_exceptions import exceptions
from pyravendb.data.indexes import IndexFieldOptions, SortOptions, IndexDefinition
from pyravendb.data.query import IndexQuery
from pyravendb.tests.test_base import TestBase
from pyravendb.raven_operations.maintenance_operations import PutIndexesOperation
from pyravendb.raven_operations.operations import QueryOperationOptions, PatchByQueryOperation, DeleteByQueryOperation


class TestByIndexActions(TestBase):
    def setUp(self):
        super(TestByIndexActions, self).setUp()
        index_map = ("from doc in docs.Testings "
                     "select new{"
                     "Name = doc.Name,"
                     "DocNumber = doc.DocNumber} ")

        self.index_sort = IndexDefinition(name="Testing_Sort", maps=index_map,
                                          fields={"DocNumber": IndexFieldOptions(sort_options=SortOptions.numeric)})
        self.patch = "this.Name = 'Patched';"
        self.store.maintenance.send(PutIndexesOperation(self.index_sort))
        self.requests_executor = self.store.get_request_executor()
        for i in range(100):
            put_command = PutDocumentCommand("testing/" + str(i),
                                             {"Name": "test" + str(i), "DocNumber": i,
                                              "@metadata": {"@collection": "Testings"}})
            self.requests_executor.execute(put_command)

    def tearDown(self):
        super(TestByIndexActions, self).tearDown()
        self.delete_all_topology_files()

    def test_update_by_index_success(self):
        query_command = QueryCommand(
            index_query=IndexQuery(query="From INDEX 'Testing_Sort'", wait_for_non_stale_results=True),
            conventions=DocumentConventions())
        self.requests_executor.execute(query_command)

        patch_command = PatchByQueryOperation("From INDEX 'Testing_Sort' Update {{{0}}}".format(
            self.patch), options=QueryOperationOptions(allow_stale=False)).get_command(self.store,
                                                                                       self.store.conventions)
        result = self.requests_executor.execute(patch_command)
        result = self.store.operations.wait_for_operation_complete(result['operation_id'])
        self.assertIsNotNone(result)
        self.assertTrue(result["Result"]["Total"] >= 50)

    def test_update_by_index_fail(self):
        patch_command = PatchByQueryOperation(
            IndexQuery("From INDEX 'TeSort' Update {{{0}}}".format(
                self.patch)), options=QueryOperationOptions(allow_stale=False)).get_command(self.store,
                                                                                            self.store.conventions)
        with self.assertRaises(exceptions.ErrorResponseException):
            self.requests_executor.execute(patch_command)

    def test_delete_by_index_fail(self):
        delete_by_index_command = DeleteByQueryOperation("From Index 'region_2' WHERE Name = 'Western'").get_command(
            self.store, self.store.conventions)
        with self.assertRaises(exceptions.ErrorResponseException):
            self.requests_executor.execute(delete_by_index_command)

    def test_delete_by_index_success(self):
        query_command = QueryCommand(self.store.conventions,
                                     IndexQuery(query="FROM INDEX 'Testing_Sort' WHERE DocNumber BETWEEN '0' AND '49'",
                                                wait_for_non_stale_results=True))
        self.requests_executor.execute(query_command)
        delete_by_index_command = DeleteByQueryOperation(
            "FROM INDEX 'Testing_Sort' WHERE DocNumber BETWEEN '0' AND '49'",
            options=QueryOperationOptions(allow_stale=False)).get_command(
            self.store, self.store.conventions)
        response = self.requests_executor.execute(delete_by_index_command)
        result = self.store.operations.wait_for_operation_complete(response["operation_id"])
        assert result["Status"] == "Completed"


if __name__ == "__main__":
    unittest.main()



from pyravendb.commands.raven_commands import GetTopologyCommand, GetStatisticsCommand
from pyravendb.connection.requests_helpers import *
from pyravendb.custom_exceptions import exceptions
from OpenSSL import crypto
from pyravendb.data.document_conventions import DocumentConventions
from threading import Lock
from pyravendb.tools.utils import Utils
from datetime import datetime, timedelta
import time
import requests
import json
import hashlib
import errno
import os
import io
import logging

logging.basicConfig(filename='requests_executor_info.log', level=logging.DEBUG)
log = logging.getLogger()

TOPOLOGY_FILES_DIR = os.path.join(os.getcwd(), "topology_files")


class RequestsExecutor(object):
    def __new__(cls, *args, **kwargs):
        instance = super(RequestsExecutor, cls).__new__(cls)
        try:
            os.makedirs(TOPOLOGY_FILES_DIR)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        return instance

    def __init__(self, database_name, certificate, conventions=None, **kwargs):
        self._database_name = database_name
        self._certificate = certificate
        self.topology_etag = kwargs.get("topology_etag", 0)
        self._last_return_response = datetime.utcnow()
        self.conventions = conventions if conventions is not None else DocumentConventions()
        self._node_selector = kwargs.get("node_selector", None)
        self._last_known_urls = None

        self.headers = {"Accept": "application/json",
                        "Raven-Client-Version": "4.0.0.0"}

        self.update_topology_lock = Lock()
        self.update_timer_lock = Lock()
        self.lock = Lock()
        self._disable_topology_updates = kwargs.get("disable_topology_updates", False)

        self._failed_nodes_timers = {}
        self.update_topology_timer = None
        self._first_topology_update = None
        self.cluster_token = None
        self.topology_nodes = None

        self._closed = False

    @property
    def certificate(self):
        return self._certificate

    @staticmethod
    def create(urls, database_name, certificate, conventions=None):
        executor = RequestsExecutor(database_name, certificate, conventions)
        executor.start_first_topology_thread(urls)
        return executor

    @staticmethod
    def create_for_single_node(url, database_name, certificate):
        topology = Topology(etag=-1, nodes=[ServerNode(url, database_name)])
        return RequestsExecutor(database_name, certificate, node_selector=NodeSelector(topology), topology_etag=-2,
                                disable_topology_updates=True)

    def start_first_topology_thread(self, urls):
        self._first_topology_update = PropagatingThread(target=self.first_topology_update, args=(urls,), daemon=True)
        self._first_topology_update.start()

    def ensure_node_selector(self):
        if self._first_topology_update and self._first_topology_update.is_alive():
            self._first_topology_update.join()

        if not self._node_selector:
            self._node_selector = NodeSelector(Topology(etag=self.topology_etag, nodes=self.topology_nodes))

    def get_preferred_node(self):
        self.ensure_node_selector()
        return self._node_selector.get_current_node()

    def execute(self, raven_command, should_retry=True):
        if not hasattr(raven_command, 'raven_command'):
            raise ValueError("Not a valid command")
        topology_update = self._first_topology_update
        if not self._disable_topology_updates:
            if topology_update is None or topology_update.is_alive():
                try:
                    if topology_update is None:
                        with self.lock:
                            if self._first_topology_update is None:
                                if self._last_known_urls is None:
                                    raise exceptions.InvalidOperationException(
                                        "No known topology and no previously known one, cannot proceed, likely a bug")
                                self.start_first_topology_thread(self._last_known_urls)
                            topology_update = self._first_topology_update
                    topology_update.join()
                except Exception as e:
                    log.debug(str(e))
                    with self.lock:
                        if self._first_topology_update == topology_update:
                            self._first_topology_update = None
                    raise

        if self._node_selector is None:
            raise exceptions.InvalidOperationException("A connection with the server could not be established",
                                                       "node_selector cannot be None, please check your connection "
                                                       "or supply a valid node_selector")
        chosen_node = self._node_selector.get_current_node()
        return self.execute_with_node(chosen_node, raven_command, should_retry)

    def execute_with_node(self, chosen_node, raven_command, should_retry):
        while True:
            raven_command.create_request(chosen_node)
            node_index = 0 if self._node_selector is None else self._node_selector.current_node_index

            with requests.session() as session:
                raven_command.headers.update(self.headers)
                if not self._disable_topology_updates:
                    raven_command.headers["Topology-Etag"] = "\"{0}\"".format(self.topology_etag)

                data = raven_command.data
                if data and not isinstance(data, io.BufferedIOBase):
                    data = json.dumps(raven_command.data, default=self.conventions.json_default_method)

                if raven_command.files:
                    data = {"data": data}

                start_time = time.time() * 1000
                end_time = None
                try:
                    response = session.request(raven_command.method, url=raven_command.url, data=data,
                                               files=raven_command.files,
                                               cert=self._certificate, headers=raven_command.headers,
                                               stream=raven_command.use_stream)
                except Exception as e:
                    end_time = time.time() * 1000
                    if not should_retry:
                        raise
                    if not self.handle_server_down(chosen_node, node_index, raven_command, e):
                        raise exceptions.AllTopologyNodesDownException(
                            "Tried to send request to all configured nodes in the topology, "
                            "all of them seem to be down or not responding.", e)

                    chosen_node = self._node_selector.get_current_node()
                    continue

                finally:
                    if not end_time:
                        end_time = time.time() * 1000
                    elapsed_time = end_time - start_time
                    chosen_node.response_time = elapsed_time

                if response is None:
                    raise ValueError("response is invalid.")

                if response.status_code == 404:
                    return raven_command.set_response(None)
                if response.status_code == 403:
                    if self._certificate is not None:
                        with open(self._certificate, 'rb') as pem:
                            cert = crypto.load_certificate(crypto.FILETYPE_PEM, pem.read())
                            name = str(cert.get_subject().get_components()[0][1])
                    raise exceptions.AuthorizationException(
                        "Forbidden access to " + chosen_node.database + "@" + chosen_node.url + ", " +
                        ("a certificate is required." if self._certificate is None
                         else name + " does not have permission to access it or is unknown.") +
                        response.request.method + " " + response.request.path_url)
                if response.status_code == 410:
                    if should_retry:
                        self.update_topology(chosen_node, True)
                        # TODO update this after build make fastest node selector
                if response.status_code == 408 or response.status_code == 502 or response.status_code == 503 or response.status_code == 504:
                    if len(raven_command.failed_nodes) == 1:
                        node = list(raven_command.failed_nodes.keys())[0]
                        database_missing = response.headers.get("Database-Missing", None)
                        if database_missing:
                            raise exceptions.DatabaseDoesNotExistException(
                                "Database " + database_missing + " does not exists")

                        raise exceptions.UnsuccessfulRequestException(node.url, raven_command.failed_nodes[node])

                    try:
                        e = response.json()["Message"]
                    except ValueError:
                        e = None
                    if self.handle_server_down(chosen_node, node_index, raven_command, e):
                        chosen_node = self._node_selector.get_current_node()
                    continue
                if response.status_code == 409:
                    # TODO: Conflict resolution
                    # current implementation is temporary
                    try:
                        response = response.json()
                        if "Error" in response:
                            raise Exception(response["Message"], response["Type"])
                    except ValueError:
                        raise response.raise_for_status()

                if "Refresh-Topology" in response.headers:
                    self.update_topology(ServerNode(chosen_node.url, self._database_name))
                self._last_return_response = datetime.utcnow()
                return raven_command.set_response(response)

    def first_topology_update(self, initial_urls):
        error_list = []
        for url in initial_urls:
            try:
                self.update_topology(ServerNode(url, self._database_name))
                self.update_topology_timer = Utils.start_a_timer(60 * 5, self.update_topology_callback, daemon=True)
                self.topology_nodes = self._node_selector.topology.nodes
                return
            except exceptions.DatabaseDoesNotExistException:
                # Will happen on all node in the cluster, so errors immediately
                self._last_known_urls = initial_urls
                raise
            except Exception as e:
                if len(initial_urls) == 0:
                    self._last_known_urls = initial_urls
                    raise exceptions.InvalidOperationException("Cannot get topology from server: " + url, e)
                error_list.append((url, e))

        # Failed to update topology trying update from cache
        for url in initial_urls:
            if self.try_load_from_cache(url):
                self.topology_nodes = self._node_selector.topology.nodes
                return

        self._last_known_urls = initial_urls
        self.raise_exceptions(error_list)

    def raise_exceptions(self, error_list):
        raise exceptions.AggregateException("Failed to retrieve database topology from all known nodes", error_list)

    def try_load_from_cache(self, url):
        server_hash = hashlib.md5(
            "{0}{1}".format(url, self._database_name).encode(
                'utf-8')).hexdigest()
        topology_file_path = os.path.join(TOPOLOGY_FILES_DIR, server_hash + ".raven-topology")
        try:
            with open(topology_file_path, 'r') as topology_file:
                json_file = json.load(topology_file)
                self._node_selector = NodeSelector(
                    Topology.convert_json_topology_to_entity(json_file))
                self.topology_etag = -2
                self.update_topology_timer = Utils.start_a_timer(60 * 5, self.update_topology_callback, daemon=True)
                return True
        except (FileNotFoundError, json.JSONDecodeError) as e:
            log.info(e)
        return False

    def update_topology(self, node, force_update=False):
        if self._closed:
            return

        if self.update_topology_lock.acquire(False):
            try:
                if self._closed:
                    return False

                command = GetTopologyCommand()
                response = self.execute_with_node(node, command, should_retry=False)

                hash_name = hashlib.md5(
                    "{0}{1}".format(node.url, node.database).encode(
                        'utf-8')).hexdigest()

                topology_file = os.path.join(TOPOLOGY_FILES_DIR, hash_name + ".raven-topology")
                try:
                    with open(topology_file, 'w') as outfile:
                        json.dump(response, outfile, ensure_ascii=False)
                except (IOError, json.JSONDecodeError):
                    pass

                topology = Topology.convert_json_topology_to_entity(response)
                if self._node_selector is None:
                    self._node_selector = NodeSelector(topology)

                elif self._node_selector.on_update_topology(topology, force_update):
                    self.cancel_all_failed_nodes_timers()

                self.topology_etag = self._node_selector.topology.etag
            finally:
                self.update_topology_lock.release()
        else:
            return False

    def handle_server_down(self, chosen_node, node_index, command, e):
        command.failed_nodes.update({chosen_node: {"url": command.url, "error": str(e), "type": type(e).__name__}})
        node_selector = self._node_selector

        if node_selector is None:
            return False

        if chosen_node not in self._failed_nodes_timers:
            node_status = NodeStatus(self, node_index, chosen_node)
            with self.update_timer_lock:
                if self._failed_nodes_timers.get(chosen_node, None) is None:
                    self._failed_nodes_timers.update({chosen_node: node_status})
                    node_status.start_timer()

        node_selector.on_failed_request(node_index)
        current_node = node_selector.get_current_node()
        if command.is_failed_with_node(current_node):
            return False
        return True

    def cancel_all_failed_nodes_timers(self):
        failed_nodes_timers = self._failed_nodes_timers
        self._failed_nodes_timers.clear()
        for _, timer in failed_nodes_timers.items():
            timer.cancel()
            timer.join()

    def check_node_status(self, node_status):
        if self._node_selector is not None:
            nodes = self._node_selector.topology.nodes
            if node_status.node_index >= len(nodes):
                return
            server_node = nodes[node_status.node_index]
            if server_node is not node_status.node:
                self.perform_health_check(server_node, node_status)

    def perform_health_check(self, node, node_status):
        command = GetStatisticsCommand(debug_tag="failure=check")
        try:
            self.execute_with_node(node, command, should_retry=False)

        except Exception as e:
            log.info("{0} is still down".format(node.cluster_tag), e)
            failed_node_timer = self._failed_nodes_timers.get(node_status.node, None)
            if failed_node_timer is not None:
                failed_node_timer.start_timer()
            return

        failed_node_timer = self._failed_nodes_timers.pop(node_status.node, None)
        if failed_node_timer:
            del failed_node_timer

        self._node_selector.restore_node_index(node_status.node_index)

    def update_topology_callback(self):
        time = datetime.utcnow()
        if time - self._last_return_response <= timedelta(minutes=5):
            return
        try:
            self.update_topology(self._node_selector.get_current_node())
        except Exception as e:
            log.info("Couldn't Update Topology from _updateTopologyTimer task", e)

    def close(self):
        if self._closed:
            return

        self._closed = True
        self.cancel_all_failed_nodes_timers()
        if self.update_topology_timer:
            self.update_topology_timer.cancel()



from pyravendb.commands.raven_commands import RavenCommand, GetStatisticsCommand
from pyravendb.tools.utils import Utils
from pyravendb.custom_exceptions import exceptions
from pyravendb.data.indexes import IndexDefinition
from abc import abstractmethod


class MaintenanceOperation(object):
    __slots__ = ['__operation']

    def __init__(self):
        self.__operation = "MaintenanceOperation"

    @property
    def operation(self):
        return self.__operation

    @abstractmethod
    def get_command(self, conventions):
        raise NotImplementedError


class DeleteIndexOperation(MaintenanceOperation):
    def __init__(self, index_name):
        super(DeleteIndexOperation, self).__init__()
        self._index_name = index_name

    def get_command(self, conventions):
        return self.DeleteIndexCommand(self._index_name)

    class _DeleteIndexCommand(RavenCommand):
        def __init__(self, index_name):
            if not index_name:
                raise ValueError("Invalid index_name")
            super(DeleteIndexOperation._DeleteIndexCommand, self).__init__(method="DELETE")
            self._index_name = index_name

        def create_request(self, server_node):
            self.url = "{0}/databases/{1}/indexes?name={2}".format(server_node.url, server_node.database,
                                                                   Utils.quote_key(self._index_name))

        def set_response(self, response):
            pass


class GetIndexOperation(MaintenanceOperation):
    def __init__(self, index_name):
        """
        @param str index_name: The name of the index
        @returns The IndexDefinition of the index
        :rtype IndexDefinition
        """
        super(GetIndexOperation, self).__init__()
        self._index_name = index_name

    def get_command(self, conventions):
        return self._GetIndexCommand(self._index_name)

    class _GetIndexCommand(RavenCommand):
        def __init__(self, index_name):
            """
           @param str index_name: Name of the index you like to get or delete
           """
            super(GetIndexOperation._GetIndexCommand, self).__init__(method="GET", is_read_request=True)
            if index_name is None:
                raise AttributeError("index_name")
            self._index_name = index_name

        def create_request(self, server_node):
            self.url = "{0}/databases/{1}/indexes?{2}".format(server_node.url, server_node.database,
                                                              "name={0}".format(
                                                                  Utils.quote_key(self._index_name, True)))

        def set_response(self, response):
            if response is None:
                return None
            data = {}
            try:
                response = response.json()["Results"]
                if len(response) > 1:
                    raise ValueError("response is Invalid")
                for key, value in response[0].items():
                    data[Utils.convert_to_snake_case(key)] = value
                return IndexDefinition(**data)

            except ValueError:
                raise response.raise_for_status()


class GetIndexNamesOperation(MaintenanceOperation):
    def __init__(self, start, page_size):
        super(GetIndexNamesOperation, self).__init__()
        self._start = start
        self._page_size = page_size

    def get_command(self, conventions):
        return self._GetIndexNamesCommand(self._start, self._page_size)

    class _GetIndexNamesCommand(RavenCommand):
        def __init__(self, start, page_size):
            super(GetIndexNamesOperation._GetIndexNamesCommand, self).__init__(method="GET", is_read_request=True)
            self._start = start
            self._page_size = page_size

        def create_request(self, server_node):
            self.url = "{0}/databases/{1}/indexes?start={2}&pageSize={3}&namesOnly=true".format(server_node.url,
                                                                                                server_node.database,
                                                                                                self._start,
                                                                                                self._page_size)

        def set_response(self, response):
            if response is None:
                raise ValueError("Invalid response")

            response = response.json()
            if "Error" in response:
                raise exceptions.ErrorResponseException(response["Error"])
            if "Results" not in response:
                raise ValueError("Invalid response")

            return response["Results"]


class PutIndexesOperation(MaintenanceOperation):
    def __init__(self, *indexes_to_add):
        if len(indexes_to_add) == 0:
            raise ValueError("Invalid indexes_to_add")

        super(PutIndexesOperation, self).__init__()
        self._indexes_to_add = indexes_to_add

    def get_command(self, conventions):
        return self._PutIndexesCommand(*self._indexes_to_add)

    class _PutIndexesCommand(RavenCommand):
        def __init__(self, *index_to_add):
            """
            @param index_to_add: Index to add to the database
            :type args of IndexDefinition
            :rtype dict (etag, transformer)
            """
            super(PutIndexesOperation._PutIndexesCommand, self).__init__(method="PUT")
            if index_to_add is None:
                raise ValueError("None indexes_to_add is not valid")

            self.indexes_to_add = []
            for index_definition in index_to_add:
                if not isinstance(index_definition, IndexDefinition):
                    raise ValueError("index_definition in indexes_to_add must be IndexDefinition type")
                if index_definition.name is None:
                    raise ValueError("None Index name is not valid")
                self.indexes_to_add.append(index_definition.to_json())

        def create_request(self, server_node):
            self.url = "{0}/databases/{1}/admin/indexes".format(server_node.url, server_node.database)
            self.data = {"Indexes": self.indexes_to_add}

        def set_response(self, response):
            try:
                response = response.json()
                if "Error" in response:
                    raise exceptions.ErrorResponseException(response["Error"])
                return response["Results"]
            except ValueError:
                response.raise_for_status()


class GetStatisticsOperation(MaintenanceOperation):
    def get_command(self, conventions):
        return GetStatisticsCommand()



class GenerateEntityIdOnTheClient(object):
    @staticmethod
    def try_set_id_on_entity(entity, key):
        if hasattr(entity, "Id"):
            entity.Id = key

    @staticmethod
    def try_get_id_from_instance(entity):
        if entity is None:
            raise ValueError("None entity is invalid")
        return getattr(entity, "Id", None)



from pyravendb.tools.utils import Utils
from abc import ABCMeta
from enum import Enum
import json
import xxhash
import sys


class EscapeQueryOptions(Enum):
    EscapeAll = 0
    AllowPostfixWildcard = 1
    AllowAllWildcards = 2
    RawQuery = 3


class QueryOperator(Enum):
    OR = "OR"
    AND = "AND"

    def __str__(self):
        return self.value


class OrderingType(Enum):
    str = 0
    long = " AS long"
    float = " AS double"
    alpha_numeric = " AS alphaNumeric"

    def __str__(self):
        return self.value


class IndexQueryBase(object):
    __metaclass__ = ABCMeta

    def __init__(self, query, query_parameters=None, start=0, page_size=None,
                 wait_for_non_stale_results=False, wait_for_non_stale_results_timeout=None, cutoff_etag=None):
        """
        @param str query: Actual query that will be performed.
        @param dict query_parameters: Parameters to the query
        @param int start:  Number of records that should be skipped.
        @param int page_size:  Maximum number of records that will be retrieved.
        @param bool wait_for_non_stale_results: True to wait in the server side to non stale result
        @param None or float cutoff_etag: Gets or sets the cutoff etag.
        """
        self._page_size = sys.maxsize if page_size is None else page_size
        self._page_size_set = False
        if page_size is not None:
            self.page_size = page_size
        self.query = query
        self.query_parameters = query_parameters
        self.start = start
        self.wait_for_non_stale_results = wait_for_non_stale_results
        self.wait_for_non_stale_results_timeout = wait_for_non_stale_results_timeout
        self.cutoff_etag = cutoff_etag

    def __str__(self):
        return self.query

    @property
    def page_size(self):
        return self._page_size

    @page_size.setter
    def page_size(self, value):
        self._page_size = value
        self._page_size_set = True

    @property
    def page_size_set(self):
        return self._page_size_set


class IndexQuery(IndexQueryBase):
    def __init__(self, query="", query_parameters=None, start=0, includes=None, show_timings=False,
                 skip_duplicate_checking=False, **kwargs):
        super(IndexQuery, self).__init__(query=query, query_parameters=query_parameters, start=start, **kwargs)
        self.allow_multiple_index_entries_for_same_document_to_result_transformer = kwargs.get(
            "allow_multiple_index_entries_for_same_document_to_result_transformer", False)
        self.includes = includes if includes is not None else []
        self.show_timings = show_timings
        self.skip_duplicate_checking = skip_duplicate_checking

    def get_custom_query_str_variables(self):
        return ""

    def to_json(self):
        data = {"Query": self.query, "CutoffEtag": self.cutoff_etag}
        if self._page_size_set and self._page_size >= 0:
            data["PageSize"] = self.page_size
        if self.wait_for_non_stale_results:
            data["WaitForNonStaleResults"] = self.wait_for_non_stale_results
        if self.start > 0:
            data["Start"] = self.start
        if self.wait_for_non_stale_results_timeout is not None:
            data["WaitForNonStaleResultsTimeout"] = Utils.timedelta_to_str(self.wait_for_non_stale_results_timeout)
        if self.allow_multiple_index_entries_for_same_document_to_result_transformer:
            data[
                "AllowMultipleIndexEntriesForSameDocumentToResultTransformer"] = \
                self.allow_multiple_index_entries_for_same_document_to_result_transformer
        if self.show_timings:
            data["ShowTimings"] = self.show_timings
        if self.skip_duplicate_checking:
            data["SkipDuplicateChecking"] = self.skip_duplicate_checking
        if len(self.includes) > 0:
            data["Includes"] = self.includes

        data["QueryParameters"] = self.query_parameters if self.query_parameters is not None else None
        return data

    def get_query_hash(self):
        query_hash = xxhash.xxh64()
        query_hash.update(self.query)
        query_hash.update(bytes(self.wait_for_non_stale_results))
        if self.wait_for_non_stale_results_timeout:
            query_hash.update(bytes(Utils.timedelta_tick(self.wait_for_non_stale_results_timeout)))
        query_hash.update(bytes(self.show_timings))
        if self.cutoff_etag:
            query_hash.update(bytes(self.cutoff_etag))
        query_hash.update(bytes(self.start))
        query_hash.update(self.page_size.to_bytes(8, byteorder='big'))
        if self.query_parameters:
            str_query_parameters = json.dumps(self.query_parameters).encode('utf-8')
            query_hash.update(str_query_parameters)
        return query_hash.intdigest()


class FacetQuery(IndexQueryBase):
    def __init__(self, query="", facet_setup_doc=None, facets=None, start=0, page_size=None, **kwargs):
        """
        @param facets: list of facets (mutually exclusive with FacetSetupDoc).
        :type list of Facet
        @param facet_setup_doc: Id of a facet setup document that can be found in database containing facets.
        :type str
        """
        super().__init__(query=query, start=start, page_size=page_size, **kwargs)
        self.facets = {} if facets is None else facets
        self.facet_setup_doc = facet_setup_doc

    def get_query_hash(self):
        query_hash = xxhash.xxh64()
        query_hash.update(self.query)
        query_hash.update(bytes(self.wait_for_non_stale_results))
        if self.wait_for_non_stale_results_timeout:
            query_hash.update(bytes(Utils.timedelta_tick(self.wait_for_non_stale_results_timeout)))
        if self.cutoff_etag:
            query_hash.update(bytes(self.cutoff_etag))
        query_hash.update(bytes(self.start))
        query_hash.update(self.page_size.to_bytes(8, byteorder='big'))
        if self.query_parameters:
            str_query_parameters = json.dumps(self.query_parameters).encode('utf-8')
            query_hash.update(str_query_parameters)
        if self.facet_setup_doc:
            query_hash.update(bytes(self.facet_setup_doc))
        if self.facets:
            json_facets = [facet.to_json() for facet in self.facets]
            str_facets = json.dumps(json_facets).encode('utf-8')
            query_hash.update(str_facets)
        return query_hash.intdigest()

    def to_json(self):
        data = {"Query": self.query, "CutoffEtag": self.cutoff_etag}
        if self._page_size_set and self._page_size >= 0:
            data["PageSize"] = self.page_size
        if self.start > 0:
            data["Start"] = self.start
        data["Facets"] = [facet.to_json() for facet in self.facets]
        data["FacetSetupDoc"] = self.facet_setup_doc
        data["QueryParameters"] = self.query_parameters if self.query_parameters is not None else None
        data["WaitForNonStaleResults"] = self.wait_for_non_stale_results
        data["WaitForNonStaleResultsTimeout"] = Utils.timedelta_to_str(self.wait_for_non_stale_results_timeout) \
            if self.wait_for_non_stale_results_timeout is not None else None
        return data


class FacetMode(Enum):
    default = "Default"
    ranges = "Ranges"

    def __str__(self):
        return self.value


class FacetAggregation(Enum):
    none = "None"
    count = "Count"
    max = "Max"
    min = "Min"
    average = "Average"
    sum = "Sum"

    def __str__(self):
        return self.value


class FacetTermSortMode(Enum):
    value_asc = "ValueAsc"
    value_desc = "ValueDesc"
    hits_asc = "HitsAsc"
    hits_desc = "HitsDesc"

    def __str__(self):
        return self.value


class Facet(object):
    def __init__(self, name=None, display_name=None, ranges=None, mode=FacetMode.default,
                 aggregation=FacetAggregation.none, aggregation_field=None, aggregation_type=None, max_result=None,
                 term_sort_mode=FacetTermSortMode.value_asc, include_remaining_terms=False):
        """
        @param name: Name of facet.
        :type str
        @param display_name: Display name of facet. Will return {name} if None.
        :type str
        @param list ranges: List of facet ranges
        ex. [10 TO NULL] is > 10, [NULL TO 10} is <=10, [10 TO 30] is 10 < x > 30
        @param mode: Mode of a facet (Default, ranges).
        :type FacetMode
        @param aggregation: Flags indicating type of facet aggregation
        :type FacetAggregation
        @param aggregation_field: Field on which aggregation will be performed
        :type str
        @param aggregation_type: Type of field on which aggregation will be performed.
        :type str
        @param max_result: Maximum number of results to return.
        :type int
        @param term_sort_mode: Indicates how terms should be sorted.
        :type FacetTermSortMode
        @param include_remaining_terms: Indicates if remaining terms should be included in results.
        :type bool
        """
        self.name = name
        self.display_name = name if display_name is None else display_name
        self.ranges = [] if ranges is None else ranges
        self.mode = mode
        self.aggregation = aggregation
        self.aggregation_field = aggregation_field
        self.aggregation_type = aggregation_type
        self.max_result = max_result
        self.term_sort_mode = term_sort_mode
        self.include_remaining_terms = include_remaining_terms

    def to_json(self):
        data = {"Mode": str(self.mode), "Aggregation": str(self.aggregation),
                "AggregationField": self.aggregation_field,
                "AggregationType": self.aggregation_type, "Name": self.name, "TermSortMode": str(self.term_sort_mode),
                "IncludeRemainingTerms": self.include_remaining_terms}

        if self.max_result is not None:
            data["MaxResults"] = self.max_result

        if self.display_name is not None and self.display_name != self.name:
            data["DisplayName"] = self.display_name

        if len(self.ranges) > 0:
            data["Ranges"] = self.ranges

        return data



import unittest
from pyravendb.commands.raven_commands import PutDocumentCommand, GetDocumentCommand
from pyravendb.tests.test_base import TestBase


class TestGet(TestBase):
    def setUp(self):
        super(TestGet, self).setUp()
        put_command = PutDocumentCommand("products/101", {"Name": "test", "@metadata": {}})
        put_command2 = PutDocumentCommand("products/10", {"Name": "test", "@metadata": {}})
        self.requests_executor = self.store.get_request_executor()
        self.requests_executor.execute(put_command)
        self.requests_executor.execute(put_command2)
        self.response = self.requests_executor.execute(GetDocumentCommand("products/101"))
        self.other_response = self.requests_executor.execute(GetDocumentCommand("products/10"))

    def tearDown(self):
        super(TestGet, self).tearDown()
        self.delete_all_topology_files()

    def test_equal(self):
        self.assertEqual(self.response["Results"][0]["@metadata"]["@id"], "products/101")

    def test_not_equal(self):
        self.assertNotEqual(self.response["Results"][0]["@metadata"]["@id"],
                            self.other_response["Results"][0]["@metadata"]["@id"])

    def test_null(self):
        self.assertIsNone(self.requests_executor.execute(GetDocumentCommand("product")))


if __name__ == "__main__":
    unittest.main()



from pyravendb.commands.raven_commands import GetClusterTopologyCommand, GetTcpInfoCommand
from pyravendb.connection.requests_executor import RequestsExecutor
from pyravendb.connection.requests_helpers import *
from pyravendb.custom_exceptions import exceptions
import hashlib
import json
import os

import logging

logging.basicConfig(filename='cluster_requests_executor_info.log', level=logging.DEBUG)
log = logging.getLogger()

TOPOLOGY_FILES_DIR = os.path.join(os.getcwd(), "topology_files")


class ClusterRequestExecutor(RequestsExecutor):
    def __new__(cls, *args, **kwargs):
        return super(ClusterRequestExecutor, cls).__new__(cls)

    def __init__(self, certificate, convention=None, **kwargs):
        super(ClusterRequestExecutor, self).__init__(None, certificate, convention, **kwargs)
        self.update_cluster_topology_lock = Lock()

    @staticmethod
    def create(urls, certificate, convention=None, **kwargs):
        executor = ClusterRequestExecutor(certificate, convention, **kwargs)
        executor.start_first_topology_thread(urls)
        return executor

    @staticmethod
    def create_for_single_node(url, certificate, convention=None):
        topology = Topology(etag=-1, nodes=[ServerNode(url)])
        return ClusterRequestExecutor(certificate, convention, node_selector=NodeSelector(topology), topology_etag=-2,
                                      disable_topology_updates=True)

    def update_topology(self, node, force_update=False):
        if self._closed:
            return

        if self.update_cluster_topology_lock.acquire(False):
            try:
                if self._closed:
                    return False

                command = GetClusterTopologyCommand()
                response = self.execute_with_node(node, command, should_retry=False)

                hash_name = hashlib.md5(
                    "{0}".format(node.url).encode(
                        'utf-8')).hexdigest()

                topology_file = os.path.join(TOPOLOGY_FILES_DIR, hash_name + ".raven-cluster-topology")
                try:
                    with open(topology_file, 'w') as outfile:
                        json.dump(response, outfile, ensure_ascii=False)
                except:
                    pass

                topology = Topology()
                for key, url in response["Topology"]["Members"].items():
                    topology.nodes.append(ServerNode(url, cluster_tag=key))
                if self._node_selector is None:
                    self._node_selector = NodeSelector(topology)

                elif self._node_selector.on_update_topology(topology, force_update):
                    self.cancel_all_failed_nodes_timers()

                self.topology_etag = self._node_selector.topology.etag
            finally:
                self.update_cluster_topology_lock.release()
        else:
            return False

    def raise_exceptions(self, error_list):
        raise exceptions.AggregateException("Failed to retrieve cluster topology from all known nodes", error_list)

    def try_load_from_cache(self, url):
        server_hash = hashlib.md5(
            "{0}{1}".format(url, self._database_name).encode(
                'utf-8')).hexdigest()
        cluster_topology_file_path = os.path.join(TOPOLOGY_FILES_DIR, server_hash + ".raven-cluster-topology")
        try:
            with open(cluster_topology_file_path, 'r') as cluster_topology_file:
                json_file = json.load(cluster_topology_file)
                topology = Topology()
                for key, url in json_file["Topology"]["Members"].items():
                    topology.nodes.append(ServerNode(url, cluster_tag=key))
                self._node_selector = NodeSelector(
                    Topology.convert_json_topology_to_entity(json.load(topology)))
                return True
        except (FileNotFoundError, json.JSONDecodeError) as e:
            log.info(e)
        return False

    def perform_health_check(self, node, node_status):
        command = GetTcpInfoCommand(tag="health-check")
        try:
            self.execute_with_node(node, command, should_retry=False)

        except Exception as e:
            log.info("{0} is still down".format(node.cluster_tag), e)
            failed_node_timer = self._failed_nodes_timers.get(node_status.node, None)
            if failed_node_timer is not None:
                failed_node_timer.start_timer()
            return

        failed_node_timer = self._failed_nodes_timers.pop(node_status.node, None)
        if failed_node_timer:
            del failed_node_timer

        self._node_selector.restore_node_index(node_status.node_index)



from pyravendb.tests.test_base import *
from pyravendb.raven_operations.operations import *
import unittest


class TestOperations(TestBase):
    def setUp(self):
        super(TestOperations, self).setUp()
        with self.store.open_session() as session:
            user = User(name="Idan", age="29")
            session.store(user, key="users/1-A")
            session.store(Dog(name="Fazy", owner=user), key="dogs/1-A")
            session.save_changes()

    def tearDown(self):
        super(TestOperations, self).tearDown()
        TestBase.delete_all_topology_files()

    def test_put_attachment(self):
        binary_list = bin(int(''.join(map(str, [1, 2, 3, 4, 5, 70, 90]))) << 1)
        put_result = self.store.operations.send(
            PutAttachmentOperation("users/1-A", "my_picture", binary_list, content_type="image/png"))
        self.assertIsNotNone(put_result)
        self.assertEqual(put_result["ContentType"], "image/png")
        self.assertEqual(put_result["Name"], "my_picture")

    def test_delete_attachment(self):
        self.test_put_attachment()
        get_attachment_operation = GetAttachmentOperation("users/1-A", "my_picture", AttachmentType.document, None)
        attachment = self.store.operations.send(get_attachment_operation)
        self.assertIsNotNone(attachment)
        self.store.operations.send(DeleteAttachmentOperation("users/1-A", "my_picture"))
        attachment = self.store.operations.send(get_attachment_operation)
        self.assertIsNone(attachment)

    def test_patch_by_index(self):
        self.store.maintenance.send(PutIndexesOperation(
            IndexDefinition("Patches", maps="from doc in docs.Patches select new {patched = doc.patched}")))

        with self.store.open_session() as session:
            session.store(Patch(patched=False))
            session.store(Patch(patched=False))
            session.store(Patch(patched=True))
            session.save_changes()
            # doing the query here to make sure the query won't be stale
            query_result = list(
                session.query(wait_for_non_stale_results=True).raw_query("FROM INDEX 'Patches' Where patched=False"))
            assert query_result

        options = QueryOperationOptions(allow_stale=False, retrieve_details=True)
        operation = PatchByQueryOperation(
            query_to_update=IndexQuery(query="FROM INDEX 'Patches' Where patched=False UPDATE {{this.patched=true}}"),
            options=options)
        operation_id = self.store.operations.send(operation)
        result = self.store.operations.wait_for_operation_complete(operation_id=operation_id["operation_id"])
        self.assertTrue(len(result["Result"]["Details"]) == 2)

    def test_fail_patch_wrong_index_name(self):
        options = QueryOperationOptions(allow_stale=False, retrieve_details=True)
        operation = PatchByQueryOperation(
            query_to_update=IndexQuery(query="FROM INDEX 'None' UPDATE {{this.name='NotExist'}}",
                                       wait_for_non_stale_results=True), options=options)
        with self.assertRaises(exceptions.ErrorResponseException):
            self.store.operations.send(operation)

    def test_delete_by_index(self):
        self.store.maintenance.send(PutIndexesOperation(
            IndexDefinition("Users", maps="from doc in docs.Users select new {name=doc.name}")))

        with self.store.open_session() as session:
            session.store(User(name="delete", age=0), key="deletes/1-A")
            session.save_changes()
            # doing the query here to make sure the query won't be stale
            query_result = list(session.query(wait_for_non_stale_results=True).raw_query("FROM INDEX 'Users'"))
            assert query_result

        index_query = IndexQuery(query="FROM INDEX 'Users' WHERE name='delete'")
        operation_id = self.store.operations.send(
            DeleteByQueryOperation(index_query))
        result = self.store.operations.wait_for_operation_complete(operation_id=operation_id["operation_id"])
        self.assertEqual(result["Result"]["Total"], 1)

    def test_patch_by_collection(self):
        with self.store.open_session() as session:
            session.store(Patch(patched=False))
            session.store(Patch(patched=False))
            session.store(Patch(patched=False))
            session.save_changes()

        index_query = IndexQuery("From Patches Update {{this.patched=true}}", wait_for_non_stale_results=True)

        operation_id = self.store.operations.send(
            PatchByQueryOperation(query_to_update=index_query))
        result = self.store.operations.wait_for_operation_complete(operation_id=operation_id["operation_id"])
        self.assertTrue(result["Result"]["Total"] == 3)

    def test_delete_by_collection_(self):
        index_query = IndexQuery("From Users")
        operation_id = self.store.operations.send(DeleteByQueryOperation(query_to_delete=index_query))["operation_id"]
        result = self.store.operations.wait_for_operation_complete(operation_id)
        self.assertTrue(result["Result"]["Total"] == 1)


if __name__ == "__main__":
    unittest.main()



from pyravendb.tests.test_base import TestBase
from datetime import datetime, timedelta
from pyravendb.tools.utils import Utils
import unittest


class Time(object):
    def __init__(self, td, dt):
        self.td = td
        self.dt = dt


class TestConversion(TestBase):
    def setUp(self):
        super(TestConversion, self).setUp()

        with self.store.open_session() as session:
            session.store(Time(Utils.timedelta_to_str(timedelta(days=20, minutes=23, seconds=59, milliseconds=254)),
                               Utils.datetime_to_string(datetime.now())), "times/3")
            session.store(Time(Utils.timedelta_to_str(timedelta(minutes=23, seconds=59, milliseconds=254)),
                               Utils.datetime_to_string(datetime.now())), "times/4")
            session.save_changes()

    def tearDown(self):
        super(TestConversion, self).tearDown()
        self.delete_all_topology_files()

    def test_load_timedelta_and_datetime(self):
        with self.store.open_session() as session:
            times = session.load("times/3", object_type=Time, nested_object_types={"td": timedelta, "dt": datetime})
            self.assertTrue(isinstance(times.td, timedelta) and isinstance(times.dt, datetime))

    def test_store_conversion(self):
        with self.store.open_session() as session:
            times = Time(timedelta(days=1, hours=1), datetime.now())
            session.store(times)
            key = session.advanced.get_document_id(times)
            session.save_changes()

        with self.store.open_session() as session:
            times = session.load(key, object_type=Time, nested_object_types={"td": timedelta, "dt": datetime})
            self.assertTrue(isinstance(times.td, timedelta) and isinstance(times.dt, datetime))

    def test_query_conversion(self):
        with self.store.open_session() as session:
            query = list(session.query(object_type=Time, nested_object_types={"td": timedelta,
                                                                              "dt": datetime}).
                         where_greater_than_or_equal("td", timedelta(days=9)))

            not_working = False
            if len(query) < 1:
                not_working = True
            for item in query:
                if not isinstance(item.td, timedelta):
                    not_working = True

            self.assertFalse(not_working)
            self.assertEqual(len(query), 1)


if __name__ == "__main__":
    unittest.main()



from pyravendb.tools.utils import Utils
from enum import Enum
from datetime import timedelta
import ijson
from _elementtree import ParseError
from socket import timeout
import re


class SubscriptionOpeningStrategy(Enum):
    """
    The client will successfully open a subscription only if there isn't any other currently connected client.
    Otherwise it will end up with SubscriptionInUseException.
    """
    open_if_free = "OpenIfFree"
    """
    The connecting client will successfully open a subscription even if there is another active subscription's consumer.
    If the new client takes over an existing client then the existing one will get a SubscriptionInUseException.

    The subscription will always be held by the last connected client.
    """
    take_over = "TakeOver"
    """
    If the client currently cannot open the subscription because it is used by another client. it will wait for that 
    client to complete and keep attempting to gain the subscription
    """
    wait_for_free = "WaitForFree"

    def __str__(self):
        return self.value


class SubscriptionCreationOptions:
    def __init__(self, query, name=None, change_vector=None):
        if not query:
            raise ValueError("Invalid subscriptions query")
        self.name = name
        self.change_vector = change_vector
        self.query = query

    def to_json(self):
        return {"ChangeVector": self.change_vector, "Query": self.query, "Name": self.name}


class SubscriptionWorkerOptions:
    def __init__(self, subscription_name, strategy=None, ignore_subscriber_errors=False,
                 time_to_wait_before_connection_retry=None, max_docs_per_batch=4096, max_erroneous_period=None,
                 close_when_no_docs_left=False):
        """
        @param str subscription_name: The subscription name
        @param SubscriptionOpeningStrategy strategy: Options for opening a subscription
        @param ignore_subscriber_errors: Ignore subscriber errors
        @param time_to_wait_before_connection_retry: The time to wait for retry
        @param int max_docs_per_batch: The max docs to send in one batch
        @param bool close_when_no_docs_left: Will continue the subscription work until the server have no more new documents to send
        """
        if not subscription_name:
            raise AttributeError("Value cannot be null or empty.", "subscription_name")
        self.subscription_name = subscription_name
        self.strategy = strategy if strategy is not None else SubscriptionOpeningStrategy.open_if_free
        self.ignore_subscriber_errors = ignore_subscriber_errors
        if time_to_wait_before_connection_retry is None:
            time_to_wait_before_connection_retry = timedelta(milliseconds=5000)
        self.time_to_wait_before_connection_retry = time_to_wait_before_connection_retry
        self.max_docs_per_batch = max_docs_per_batch
        self.max_erroneous_period = max_erroneous_period
        if max_erroneous_period is None:
            self.max_erroneous_period = timedelta(minutes=5)
        self.close_when_no_docs_left = close_when_no_docs_left

    def to_json(self):
        return {"SubscriptionName": self.subscription_name,
                "TimeToWaitBeforeConnectionRetry": str(self.time_to_wait_before_connection_retry),
                "Strategy": str(self.strategy), "MaxDocsPerBatch": self.max_docs_per_batch,
                "IgnoreSubscriberErrors": self.ignore_subscriber_errors,
                "CloseWhenNoDocsLeft": self.close_when_no_docs_left}


class SubscriptionState:
    def __init__(self, query=None, change_vector_for_next_batch_starting_point=None, subscription_id=0,
                 subscription_name=None, last_time_server_made_progress_with_documents=None,
                 last_client_connection_time=None, disabled=False, **kwargs):
        self.query = query
        self.change_vector_for_next_batch_starting_point = change_vector_for_next_batch_starting_point
        self.subscription_id = subscription_id
        self.subscription_name = subscription_name
        if isinstance(last_time_server_made_progress_with_documents, str):
            last_time_server_made_progress_with_documents = Utils.string_to_datetime(
                last_time_server_made_progress_with_documents)
        self.last_time_server_made_progress_with_documents = last_time_server_made_progress_with_documents
        if isinstance(last_client_connection_time, str):
            last_client_connection_time = Utils.string_to_datetime(last_client_connection_time)
        self.last_client_connection_time = last_client_connection_time
        self.closed = disabled
        self.latest_change_vector_client_acknowledged = kwargs.get("LatestChangeVectorClientACKnowledged", None)

    def to_json(self):
        return {"Query": self.query,
                "ChangeVectorForNextBatchStartingPoint": self.change_vector_for_next_batch_starting_point,
                "SubscriptionId": self.subscription_id, "SubscriptionName": self.subscription_name,
                "LastTimeServerMadeProgressWithDocuments": Utils.datetime_to_string(
                    self.last_time_server_made_progress_with_documents),
                "LastClientConnectionTime": Utils.datetime_to_string(self.last_client_connection_time),
                "Disabled": self.closed}


LEXEME_RE = re.compile(b'[a-z0-9eE\.\+-]+|\S')
BYTE_ARRAY_CHARACTERS = bytearray(b',}:{"')
IS_WEBSOCKET = False


# The code imported from ijson to be able to receive json from socket
class IncrementalJsonParser:
    def __init__(self, socket, is_websocket=False):
        global IS_WEBSOCKET
        IS_WEBSOCKET = is_websocket
        self.lexer = self.lexer(socket)

    @staticmethod
    def lexer(socket, buf_size=ijson.backend.BUFSIZE):
        data = socket.recv() if IS_WEBSOCKET else socket.recv(buf_size)
        if not data:
            return

        buf = data[1:-1].encode('utf-8') if IS_WEBSOCKET else data
        pos = 0
        discarded = 0
        while True:
            match = LEXEME_RE.search(buf, pos)
            if pos < len(buf) and match:
                lexeme = match.group()
                if lexeme == b'"':
                    pos = match.start()
                    start = pos + 1
                    while True:
                        try:
                            end = buf.index(b'"', start)
                            escpos = end - 1
                            while buf[escpos] == '\\':
                                escpos -= 1
                            if (end - escpos) % 2 == 0:
                                start = end + 1
                            else:
                                break
                        except ValueError:
                            data = socket.recv().encode('utf-8') if IS_WEBSOCKET else socket.recv(buf_size)
                            if not data:
                                raise ijson.backend.common.IncompleteJSONError('Incomplete string lexeme')

                            buf += data[1:-1] if IS_WEBSOCKET else data

                    yield discarded + pos, buf[pos:end + 1].decode('utf-8')
                    pos = end + 1
                else:
                    while match.end() == len(buf) and buf[pos] not in BYTE_ARRAY_CHARACTERS:
                        try:
                            data = socket.recv().encode('utf-8') if IS_WEBSOCKET else socket.recv(buf_size)
                            if not data:
                                break
                            buf += data[1:-1].encode('utf-8') if IS_WEBSOCKET else data
                            match = LEXEME_RE.search(buf, pos)
                            lexeme = match.group()
                        except timeout:
                            break

                    yield match.start(), lexeme.decode('utf-8')
                    pos = match.end()
            else:
                data = socket.recv().encode('utf-8') if IS_WEBSOCKET else socket.recv(buf_size)
                if not data:
                    break
                discarded += len(buf)
                buf = data[1:-1] if IS_WEBSOCKET else data
                pos = 0

    def create_array(self, gen):
        arr = []

        for (token, val) in gen:
            if token == "end_array":
                return arr
            arr.append(self.get_value(gen))

        raise ParseError("End array expected, but the generator ended before we got it")

    def get_value(self, gen):
        (token, val) = next(gen)

        if token == "start_array":
            return self.create_array(gen)

        if token == "start_map":
            return self.create_object(gen)

        return val

    def create_object(self, gen):
        obj = {}

        for (token, val) in gen:
            if token == "end_map":
                return obj
            if token == "map_key":
                obj[val] = self.get_value(gen)

        raise ParseError("End object expected, but the generator ended before we got it")

    def next_object(self):

        try:
            (_, text) = next(self.lexer)
            if IS_WEBSOCKET and text == ',':
                (_, text) = next(self.lexer)
        except StopIteration:
            return None

        if text != '{':
            raise ParseError("Expected start object, got: " + text)

        gen = IncrementalJsonParser.parse_object(self.lexer)
        (token, val) = next(gen)
        assert token == "start_map"

        return self.create_object(gen)

    @staticmethod
    def parse_value(lexer, symbol=None, pos=0):
        try:
            if symbol is None:
                pos, symbol = next(lexer)
            if symbol == 'null':
                yield ('null', None)
            elif symbol == 'true':
                yield ('boolean', True)
            elif symbol == 'false':
                yield ('boolean', False)
            elif symbol == '[':
                for event in IncrementalJsonParser.parse_array(lexer):
                    yield event
            elif symbol == '{':
                for event in IncrementalJsonParser.parse_object(lexer):
                    yield event
            elif symbol[0] == '"':
                yield ('string', IncrementalJsonParser.unescape(symbol[1:-1]))
            else:
                # if we got a partial token for false / null / true we need to read from the network again
                while symbol[0] in ('t', 'n') and len(symbol) < 4 or symbol[0] == 'f' and len(symbol) < 5:
                    _, nextpart = next(lexer)
                    if symbol == 'null':
                        yield ('null', None)
                    elif symbol == 'true':
                        yield ('boolean', True)
                    elif symbol == 'false':
                        yield ('boolean', False)
                    return

                try:
                    yield ('number', ijson.backend.common.number(symbol))
                except ijson.backend.decimal.InvalidOperation:
                    raise ijson.backend.UnexpectedSymbol(symbol, pos)
        except StopIteration:
            raise ijson.backend.common.IncompleteJSONError('Incomplete JSON data')

    @staticmethod
    def parse_array(lexer):
        yield ('start_array', None)
        try:
            pos, symbol = next(lexer)
            if symbol != ']':
                while True:
                    for event in IncrementalJsonParser.parse_value(lexer, symbol, pos):
                        yield event
                    pos, symbol = next(lexer)
                    if symbol == ']':
                        break
                    if symbol != ',':
                        raise ijson.backend.UnexpectedSymbol(symbol, pos)
                    pos, symbol = next(lexer)
            yield ('end_array', None)
        except StopIteration:
            raise ijson.backend.common.IncompleteJSONError('Incomplete JSON data')

    @staticmethod
    def parse_object(lexer):
        yield ('start_map', None)
        try:
            pos, symbol = next(lexer)
            if symbol != '}':
                while True:
                    if symbol[0] != '"':
                        raise ijson.backend.UnexpectedSymbol(symbol, pos)
                    yield ('map_key', IncrementalJsonParser.unescape(symbol[1:-1]))
                    pos, symbol = next(lexer)
                    if symbol != ':':
                        raise ijson.backend.UnexpectedSymbol(symbol, pos)
                    for event in IncrementalJsonParser.parse_value(lexer, None, pos):
                        yield event
                    pos, symbol = next(lexer)
                    if symbol == '}':
                        break
                    if symbol != ',':
                        raise ijson.backend.UnexpectedSymbol(symbol, pos)
                    pos, symbol = next(lexer)
            yield ('end_map', None)
        except StopIteration:
            raise ijson.backend.common.IncompleteJSONError('Incomplete JSON data')

    @staticmethod
    def unescape(s):
        start = 0
        result = ''
        while start < len(s):
            pos = s.find('\\', start)
            if pos == -1:
                if start == 0:
                    return s
                result += s[start:]
                break
            result += s[start:pos]
            pos += 1
            esc = s[pos]
            if esc == 'u':
                result += chr(int(s[pos + 1:pos + 5], 16))
                pos += 4
            elif esc == 'b':
                result += '\b'
            elif esc == 'f':
                result += '\f'
            elif esc == 'n':
                result += '\n'
            elif esc == 'r':
                result += '\r'
            elif esc == 't':
                result += '\t'
            else:
                result += esc
            start = pos + 1
        return result



from enum import Enum


class CertificateDefinition:
    def __init__(self, certificate, server_admin=False, thumbprint=None, permissions=None):
        """
        @param certificate: X509 byte array object encoded to base64
        :type str
        @param server_admin:  A part of the permissions if True get all the permissions (admin use only)
        :type bool
        @param thumbprint: The thump print of the certificate file
        :type str
        @param permissions: A dict of database_access
        :type dict[str:DatabaseAccess]
        """
        self.certificate = certificate
        self.server_admin = server_admin
        self.thumbprint = thumbprint
        self.permissions = {} if permissions is None else permissions

    def to_json(self):
        for key, value in self.permissions.items():
            self.permissions[key] = str(value)
        return {"Certificate": self.certificate, "Thumbprint": self.thumbprint, "Permissions": self.permissions}


class DatabaseAccess(Enum):
    read_write = "ReadWrite"
    admin = "Admin"

    def __str__(self):
        return self.value


class SecurityClearance(Enum):
    unauthenticated_clients = "UnauthenticatedClients"
    cluster_admin = "ClusterAdmin"
    operator = "Operator"
    valid_user = "ValidUser"

    def __str__(self):
        return self.value



from pyravendb.tests.test_base import TestBase
from pyravendb.data.indexes import IndexDefinition, FieldIndexing, IndexFieldOptions
from pyravendb.data.query import QueryOperator
from pyravendb.raven_operations.maintenance_operations import PutIndexesOperation
from datetime import datetime


class LastFm(object):
    def __init__(self, artist="", track_id="", title="", datetime_time=None, tags=None):
        self.artist = artist
        self.track_id = track_id
        self.title = title
        self.datetime_time = datetime_time if datetime_time is not None else datetime.now()
        self.tags = tags if tags is not None else []


class LastFmAnalyzed(object):
    class Result(object):
        def __init__(self, query):
            self.query = query

    def __init__(self):
        index_map = ("from song in docs.LastFms "
                     "select new {"
                     "query = new object[] {"
                     "song.artist,"
                     "((object)song.datetime_time),"
                     "song.tags,"
                     "song.title,"
                     "song.track_id}}")

        self.index_definition = IndexDefinition(name=LastFmAnalyzed.__name__, maps=index_map,
                                                fields={"query": IndexFieldOptions(indexing=FieldIndexing.search)})

    def execute(self, store):
        store.maintenance.send(PutIndexesOperation(self.index_definition))


class FullTextSearchTest(TestBase):
    def setUp(self):
        super(FullTextSearchTest, self).setUp()
        LastFmAnalyzed().execute(self.store)
        with self.store.open_session() as session:
            session.store(LastFm("Tania Maria", "TRALPJJ128F9311763", "Come With Me", datetime.now()), "LastFms/1")
            session.store(LastFm("Meghan Trainor", "TRBCNGI128F42597B4", "Me Too", datetime.now()), "LastFms/2")
            session.store(LastFm("Willie Bobo", "TRAACNS128F14A2DF5", "Spanish Grease", datetime.now()), "LastFms/3")
            session.save_changes()

    def tearDown(self):
        super(FullTextSearchTest, self).tearDown()
        self.delete_all_topology_files()

    def test_full_text_search_one(self):
        with self.store.open_session() as session:
            query = list(session.query(object_type=LastFm, wait_for_non_stale_results=True,
                                       index_name=LastFmAnalyzed.__name__).search("query", search_terms="Me"))
            self.assertTrue("Me" in str(query[0].title) and "Me" in str(query[1].title))

    def test_full_text_search_two(self):
        with self.store.open_session() as session:
            query = list(session.query(object_type=LastFm, wait_for_non_stale_results=True,
                                       index_name=LastFmAnalyzed.__name__).search("query", search_terms="Me").
                         search("query", search_terms="Bobo"))
            self.assertEqual(len(query), 3)

    def test_full_text_search_with_boost(self):
        with self.store.open_session() as session:
            query = list(session.query(object_type=LastFm, wait_for_non_stale_results=True,
                                       index_name=LastFmAnalyzed.__name__).search("query", search_terms="Me").boost(10).
                         search("query", search_terms="Bobo").boost(2))
            self.assertTrue("Me" in str(query[0].title) and "Me" in str(query[1].title) and str(
                query[2].title) == "Spanish Grease")

            query = list(session.query(object_type=LastFm, wait_for_non_stale_results=True,
                                       index_name=LastFmAnalyzed.__name__).search("query", search_terms="Me").boost(2).
                         search("query", search_terms="Bobo").boost(10))
            self.assertTrue("Me" in str(query[1].title) and "Me" in str(query[2].title) and str(
                query[0].title) == "Spanish Grease")

    def test_full_text_search_with_and_operator(self):
        with self.store.open_session() as session:
            query = list(session.query(object_type=LastFm, wait_for_non_stale_results=True,
                                       index_name=LastFmAnalyzed.__name__).search("query", search_terms="Me Come",
                                                                                  operator=QueryOperator.AND))
            assert len(query) == 1
            self.assertEqual(query[0].title, "Come With Me")



from pyravendb.connection.requests_executor import RequestsExecutor
from pyravendb.custom_exceptions import exceptions
from pyravendb.data.document_conventions import DocumentConventions
from pyravendb.hilo.hilo_generator import MultiDatabaseHiLoKeyGenerator
from pyravendb.store.document_session import DocumentSession
from pyravendb.changes.database_changes import DatabaseChanges
from pyravendb.subscriptions.document_subscriptions import DocumentSubscriptions
from threading import Lock
import uuid
import time


class DocumentStore(object):
    def __init__(self, urls=None, database=None, certificate=None):
        if not isinstance(urls, list):
            urls = [urls]
        self.urls = urls
        self.database = database
        self.conventions = DocumentConventions()
        self._certificate = certificate
        self._request_executors = {}
        self._initialize = False
        self.generator = None
        self._operations_executor = None
        self.lock = Lock()
        self.add_change_lock = Lock()
        self._maintenance_operation_executor = None
        self.subscriptions = DocumentSubscriptions(self)
        self._database_changes = {}

    @property
    def certificate(self):
        return self._certificate

    @property
    def maintenance(self):
        self._assert_initialize()
        if not self._maintenance_operation_executor:
            self._maintenance_operation_executor = MaintenanceOperationExecutor(self, self.database)
        return self._maintenance_operation_executor

    @property
    def operations(self):
        self._assert_initialize()
        if not self._operations_executor:
            self._operations_executor = OperationExecutor(self)
        return self._operations_executor

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if self.generator is not None:
            self.generator.return_unused_range()

        if self.subscriptions:
            self.subscriptions.close()

        database_changes_keys = list(self._database_changes.keys())
        for key in database_changes_keys:
            self._database_changes[key].close()

        for _, request_executor in self._request_executors.items():
            request_executor.close()

    def changes(self, database=None, on_error=None, executor=None):
        self._assert_initialize()
        if not database:
            database = self.database
        with self.add_change_lock:
            if database not in self._database_changes:
                self._database_changes[database] = DatabaseChanges(request_executor=self.get_request_executor(database),
                                                                   database_name=database,
                                                                   on_close=self._on_close_change, on_error=on_error,
                                                                   executor=executor)
            return self._database_changes[database]

    def _on_close_change(self, database):
        del self._database_changes[database]

    def get_request_executor(self, db_name=None):
        self._assert_initialize()

        if db_name is None:
            db_name = self.database

        with self.lock:
            if db_name not in self._request_executors:
                self._request_executors[db_name] = RequestsExecutor.create(self.urls, db_name, self._certificate,
                                                                           self.conventions)
        return self._request_executors[db_name]

    def initialize(self):
        if not self._initialize:
            if self.urls is None or len(self.urls) == 0:
                raise ValueError("Document store URLs cannot be empty", "urls")
            if self.database is None:
                raise exceptions.InvalidOperationException("None database is not valid")

            self.generator = MultiDatabaseHiLoKeyGenerator(self)
            self._initialize = True

    def _assert_initialize(self):
        if not self._initialize:
            raise exceptions.InvalidOperationException(
                "You cannot open a session or access the database commands before initializing the document store.\
                Did you forget calling initialize()?")

    def open_session(self, database=None, request_executor=None):
        self._assert_initialize()
        session_id = uuid.uuid4()
        requests_executor = self.get_request_executor(database) if request_executor is None else request_executor
        return DocumentSession(database, self, requests_executor, session_id)

    def generate_id(self, db_name, entity):
        if self.generator:
            return self.generator.generate_document_key(db_name, entity)


# ------------------------------Operation executors ---------------------------->
class MaintenanceOperationExecutor:
    def __init__(self, document_store, database_name=None):
        self._store = document_store
        self._database_name = database_name if database_name is not None else document_store.database
        self._server_operation_executor = None
        self._request_executor = None

    @property
    def server(self):
        if not self._server_operation_executor:
            self._server_operation_executor = ServerOperationExecutor(self._store)
        return self._server_operation_executor

    @property
    def request_executor(self):
        if self._request_executor is None:
            self._request_executor = self._store.get_request_executor(self._database_name)
        return self._request_executor

    def send(self, operation):
        try:
            operation_type = getattr(operation, 'operation')
            if operation_type != "MaintenanceOperation":
                raise ValueError("operation type cannot be {0} need to be Operation".format(operation_type))
        except AttributeError:
            raise ValueError("Invalid operation")

        command = operation.get_command(self.request_executor.conventions)
        return self.request_executor.execute(command)


class ServerOperationExecutor:
    def __init__(self, document_store):
        self._store = document_store
        self._request_executor = None

    @property
    def request_executor(self):
        if self._request_executor is None:
            from pyravendb.connection.cluster_requests_executor import ClusterRequestExecutor
            if self._store.conventions.disable_topology_update:
                self._request_executor = ClusterRequestExecutor.create_for_single_node(self._store.urls[0],
                                                                                       self._store.certificate)
            else:
                self._request_executor = ClusterRequestExecutor.create(self._store.urls, self._store.certificate)
        return self._request_executor

    def send(self, operation):
        try:
            operation_type = getattr(operation, 'operation')
            if operation_type != "ServerOperation":
                raise ValueError("operation type cannot be {0} need to be Operation".format(operation_type))
        except AttributeError:
            raise ValueError("Invalid operation")

        command = operation.get_command(self.request_executor.conventions)
        return self.request_executor.execute(command)


class OperationExecutor(object):
    def __init__(self, document_store, database_name=None):
        self._store = document_store
        self._database_name = database_name if database_name is not None else document_store.database
        self._request_executor = None

    @property
    def request_executor(self):
        if self._request_executor is None:
            self._request_executor = self._store.get_request_executor(self._database_name)
        return self._request_executor

    def wait_for_operation_complete(self, operation_id, timeout=None):
        from pyravendb.commands.raven_commands import GetOperationStateCommand
        start_time = time.time()
        try:
            get_operation_command = GetOperationStateCommand(operation_id)
            while True:
                response = self.request_executor.execute(get_operation_command)
                if "Error" in response:
                    raise ValueError(response["Error"])
                if timeout and time.time() - start_time > timeout:
                    raise exceptions.TimeoutException("The Operation did not finish before the timeout end")
                if response["Status"] == "Completed":
                    return response
                if response["Status"] == "Faulted":
                    raise exceptions.InvalidOperationException(response["Result"]["Error"])
                time.sleep(0.5)
        except ValueError as e:
            raise exceptions.InvalidOperationException(e)

    def send(self, operation):
        try:
            operation_type = getattr(operation, 'operation')
            if operation_type != "Operation":
                raise ValueError("operation type cannot be {0} need to be Operation".format(operation_type))
        except AttributeError:
            raise ValueError("Invalid operation")

        command = operation.get_command(self._store, self.request_executor.conventions)
        return self.request_executor.execute(command)



from pyravendb.custom_exceptions import exceptions
from pyravendb.commands.raven_commands import RavenCommand
from pyravendb.tools.utils import Utils
from datetime import datetime
from threading import Lock


class RangeValue(object):
    def __init__(self, min_id=1, max_id=0):
        self.min_id = min_id
        self.max_id = max_id
        self.current = self.min_id - 1


class HiLoReturnCommand(RavenCommand):
    def __init__(self, tag, last, end):
        super(HiLoReturnCommand, self).__init__(method="PUT")
        self.tag = tag
        self.last = last
        self.end = end

    def create_request(self, server_node):
        path = "hilo/return?tag={0}&end={1}&last={2}".format(self.tag, self.end, self.last)
        self.url = "{0}/databases/{1}/{2}".format(server_node.url, server_node.database, path)

    def set_response(self, response):
        pass


class NextHiLoCommand(RavenCommand):
    def __init__(self, tag, last_batch_size, last_range_at, identity_parts_separator, last_range_max):
        super(NextHiLoCommand, self).__init__(method="GET")
        self.tag = tag
        self.last_batch_size = last_batch_size
        self.last_range_at = last_range_at
        self.identity_parts_separator = identity_parts_separator
        self.last_range_max = last_range_max
        self.server_tag = None

    def create_request(self, server_node):
        path = "hilo/next?tag={0}&lastBatchSize={1}&lastRangeAt={2}&identityPartsSeparator={3}&lastMax={4}".format(
            self.tag, self.last_batch_size, self.last_range_at, self.identity_parts_separator, self.last_range_max)
        self.url = "{0}/databases/{1}/{2}".format(server_node.url, server_node.database, path)

    def set_response(self, response):
        if response is None:
            raise ValueError("response is invalid.")

        if response.status_code == 201:
            response = response.json()
            return {"prefix": response["Prefix"], "server_tag": response["ServerTag"], "low": response["Low"],
                    "high": response["High"],
                    "last_size": response["LastSize"],
                    "last_range_at": response["LastRangeAt"]}
        if response.status_code == 500:
            raise exceptions.DatabaseDoesNotExistException(response.json()["Error"])
        if response.status_code == 409:
            raise exceptions.FetchConcurrencyException(response.json()["Error"])

        raise exceptions.ErrorResponseException("Something is wrong with the request")


class MultiDatabaseHiLoKeyGenerator(object):
    def __init__(self, store):
        self._store = store
        self.conventions = self._store.conventions
        self._generators = {}

    def generate_document_key(self, db_name, entity):
        db = db_name if db_name is not None else self._store.database
        if db not in self._generators:
            generator = MultiTypeHiLoKeyGenerator(self._store, db)
            self._generators[db] = generator
        else:
            generator = self._generators[db]

        return generator.generate_document_key(entity)

    def return_unused_range(self):
        for key in self._generators:
            self._generators[key].return_unused_range()


class MultiTypeHiLoKeyGenerator(object):
    def __init__(self, store, db_name):
        self._store = store
        self._db_name = db_name
        self._conventions = store.conventions
        self.lock = Lock()
        self.key_generators_by_tag = {}

    def generate_document_key(self, entity):
        tag = self._conventions.default_transform_type_tag_name(entity.__class__.__name__)
        if tag is None:
            return None
        else:
            with self.lock:
                if tag in self.key_generators_by_tag:
                    value = self.key_generators_by_tag[tag]
                else:
                    value = HiLoKeyGenerator(tag, self._store, self._db_name)
                    self.key_generators_by_tag[tag] = value
        return value.generate_document_key()

    def return_unused_range(self):
        for key in self.key_generators_by_tag:
            self.key_generators_by_tag[key].return_unused_range()


class HiLoKeyGenerator(object):
    def __init__(self, tag, store, db_name):
        self._tag = tag
        self._store = store
        self._db_name = db_name if db_name is not None else store.database
        self._last_range_at = datetime(1, 1, 1)
        self._last_batch_size = 0
        self._range = RangeValue()
        self._prefix = None
        self._server_tag = None
        self.conventions = store.conventions
        self._identity_parts_separator = self.conventions.identity_parts_separator
        self.collection_ranges = {}
        self.lock = Lock()

    def get_document_key_from_id(self, next_id):
        return "{0}{1}-{2}".format(self._prefix, next_id, self._server_tag)

    def generate_document_key(self):
        ranges = {self._tag: RangeValue()}
        ranges.update(self.collection_ranges)
        self.collection_ranges = ranges
        self.next_id()
        return self.get_document_key_from_id(self.collection_ranges[self._tag].current)

    def next_id(self):
        while True:
            my_collection_range = self.collection_ranges[self._tag]
            current_max = my_collection_range.max_id
            with self.lock:
                my_collection_range.current += 1
                current_id = my_collection_range.current
            if current_id <= current_max:
                return
            with self.lock:
                if my_collection_range != self.collection_ranges[self._tag]:
                    continue
                try:
                    self.get_next_range()
                    self.collection_ranges[self._tag] = self._range
                except exceptions.FetchConcurrencyException:
                    pass

    def get_next_range(self):
        hilo_command = NextHiLoCommand(self._tag, self._last_batch_size, self._last_range_at,
                                       self._identity_parts_separator, self._range.max_id)
        result = self._store.get_request_executor().execute(hilo_command)
        self._prefix = result["prefix"]
        self._server_tag = result["server_tag"]
        self._last_range_at = Utils.string_to_datetime(result["last_range_at"])
        self._last_batch_size = result["last_size"]
        self._range = RangeValue(result["low"], result["high"])

    def return_unused_range(self):
        return_command = HiLoReturnCommand(self._tag, self._range.current, self._range.max_id)
        self._store.get_request_executor().execute(return_command)



from pyravendb.commands.raven_commands import *
from pyravendb.data.patches import PatchRequest
from abc import abstractmethod
from pyravendb.tools.utils import Utils
from pyravendb.data.operation import AttachmentType
from pyravendb.custom_exceptions import exceptions
from pyravendb.data.query import IndexQuery


class QueryOperationOptions(object):
    """
      @param allow_stale Indicates whether operations are allowed on stale indexes.
      :type bool
      @param stale_timeout: If AllowStale is set to false and index is stale, then this is the maximum timeout to wait
      for index to become non-stale. If timeout is exceeded then exception is thrown.
      None by default - throw immediately if index is stale.
      :type timedelta
      @param max_ops_per_sec Limits the amount of base Operation per second allowed.
      :type int
      @param retrieve_details Determines whether Operation details about each document should be returned by server.
      :type bool
  """

    def __init__(self, allow_stale=True, stale_timeout=None, max_ops_per_sec=None, retrieve_details=False):
        self.allow_stale = allow_stale
        self.stale_timeout = stale_timeout
        self.retrieve_details = retrieve_details
        self.max_ops_per_sec = max_ops_per_sec


class Operation(object):
    __slots__ = ['__operation']

    def __init__(self):
        self.__operation = "Operation"

    @property
    def operation(self):
        return self.__operation

    @abstractmethod
    def get_command(self, store, conventions, cache=None):
        raise NotImplementedError


class DeleteAttachmentOperation(Operation):
    def __init__(self, document_id, name, change_vector=None):
        super(DeleteAttachmentOperation, self).__init__()
        self._document_id = document_id
        self._name = name
        self._change_vector = change_vector

    def get_command(self, store, conventions, cache=None):
        return self._DeleteAttachmentCommand(self._document_id, self._name, self._change_vector)

    class _DeleteAttachmentCommand(RavenCommand):
        def __init__(self, document_id, name, change_vector=None):
            super(DeleteAttachmentOperation._DeleteAttachmentCommand, self).__init__(method="DELETE")

            if not document_id:
                raise ValueError("Invalid document_id")
            if not name:
                raise ValueError("Invalid name")

            self._document_id = document_id
            self.name = name
            self._change_vector = change_vector

        def create_request(self, server_node):
            self.url = "{0}/databases/{1}/attachments?id={2}&name={3}".format(server_node.url, server_node.database,
                                                                              Utils.quote_key(self._document_id),
                                                                              Utils.quote_key(self.name))
            if self._change_vector is not None:
                self.headers = {"If-Match": "\"{0}\"".format(self._change_vector)}

        def set_response(self, response):
            pass


class PatchByQueryOperation(Operation):
    def __init__(self, query_to_update, options=None):
        """
        @param query_to_update: query that will be performed
        :type IndexQuery or str
        @param options: various Operation options e.g. AllowStale or MaxOpsPerSec
        :type QueryOperationOptions
        @return: json
        :rtype: dict of operation_id
        """
        if query_to_update is None:
            raise ValueError("Invalid query")
        super(PatchByQueryOperation, self).__init__()
        if isinstance(query_to_update, str):
            query_to_update = IndexQuery(query=query_to_update)
        self._query_to_update = query_to_update
        if options is None:
            options = QueryOperationOptions()
        self._options = options

    def get_command(self, store, conventions, cache=None):
        return self._PatchByQueryCommand(self._query_to_update, self._options)

    class _PatchByQueryCommand(RavenCommand):
        def __init__(self, query_to_update, options):
            """
            @param query_to_update: query that will be performed
            :type IndexQuery
            @param options: various Operation options e.g. AllowStale or MaxOpsPerSec
            :type QueryOperationOptions
            @return: json
            :rtype: dict of operation_id
            """
            super(PatchByQueryOperation._PatchByQueryCommand, self).__init__(method="PATCH")
            self._query_to_update = query_to_update
            self._options = options

        def create_request(self, server_node):
            if not isinstance(self._query_to_update, IndexQuery):
                raise ValueError("query must be IndexQuery Type")

            self.url = server_node.url + "/databases/" + server_node.database + "/queries"
            path = "?allowStale={0}&maxOpsPerSec={1}&details={2}".format(self._options.allow_stale,
                                                                         "" if self._options.max_ops_per_sec is None
                                                                         else self._options.max_ops_per_sec,
                                                                         self._options.retrieve_details)
            if self._options.stale_timeout is not None:
                path += "&staleTimeout=" + str(self._options.stale_timeout)

            self.url += path
            self.data = {"Query": self._query_to_update.to_json()}

        def set_response(self, response):
            if response is None:
                raise exceptions.ErrorResponseException("Invalid Response")
            try:
                response = response.json()
                if "Error" in response:
                    raise exceptions.ErrorResponseException(response["Error"])
                return {"operation_id": response["OperationId"]}
            except ValueError:
                raise response.raise_for_status()


class DeleteByQueryOperation(Operation):
    def __init__(self, query_to_delete, options=None):
        """
        @param query_to_delete: query that will be performed
        :type IndexQuery or str
        @param options: various Operation options e.g. AllowStale or MaxOpsPerSec
        :type QueryOperationOptions
        :rtype: dict of operation_id
       """
        if not query_to_delete:
            raise ValueError("Invalid query")

        super(DeleteByQueryOperation, self).__init__()
        if isinstance(query_to_delete, str):
            query_to_delete = IndexQuery(query=query_to_delete)
        self._query_to_delete = query_to_delete
        self._options = options if options is not None else QueryOperationOptions()

    def get_command(self, store, conventions, cache=None):
        return self._DeleteByQueryCommand(self._query_to_delete, self._options)

    class _DeleteByQueryCommand(RavenCommand):
        def __init__(self, query_to_delete, options=None):
            """
            @param query_to_delete: query that will be performed
            :type IndexQuery
            @param options: various Operation options e.g. AllowStale or MaxOpsPerSec
            :type QueryOperationOptions
            @return: json
            :rtype: dict
            """
            super(DeleteByQueryOperation._DeleteByQueryCommand, self).__init__(method="DELETE")
            self._query_to_delete = query_to_delete
            self._options = options

        def create_request(self, server_node):
            self.url = server_node.url + "/databases/" + server_node.database + "/queries"
            path = "?allowStale={0}&maxOpsPerSec={1}&details={2}".format(self._options.allow_stale,
                                                                         "" if self._options.max_ops_per_sec is None else self._options.max_ops_per_sec,
                                                                         self._options.retrieve_details)
            if self._options.stale_timeout is not None:
                path += "&staleTimeout=" + str(self._options.stale_timeout)

            self.url += path
            self.data = self._query_to_delete.to_json()

        def set_response(self, response):
            if response is None:
                raise exceptions.ErrorResponseException("Could not find index {0}".format(self.index_name))

            if response.status_code != 200 and response.status_code != 202:
                try:
                    raise exceptions.ErrorResponseException(response.json()["Error"])
                except ValueError:
                    raise response.raise_for_status()
            return {"operation_id": response.json()["OperationId"]}


class GetAttachmentOperation(Operation):
    def __init__(self, document_id, name, attachment_type, change_vector):
        if document_id is None:
            raise ValueError("Invalid document_id")
        if name is None:
            raise ValueError("Invalid name")

        if attachment_type != AttachmentType.document and change_vector is None:
            raise ValueError("change_vector",
                             "Change Vector cannot be null for attachment type {0}".format(attachment_type))

        super(GetAttachmentOperation, self).__init__()
        self._document_id = document_id
        self._name = name
        self._attachment_type = attachment_type
        self._change_vector = change_vector

    def get_command(self, store, conventions, cache=None):
        return self._GetAttachmentCommand(self._document_id, self._name, self._attachment_type, self._change_vector)

    class _GetAttachmentCommand(RavenCommand):
        def __init__(self, document_id, name, attachment_type, change_vector):
            """
            @param document_id: The id of the document
             :type str
            @param name: The name of the attachment
            :type str
            @param attachment_type: The type of the attachment
            :type AttachmentType
            @param change_vector: The change vector of the document (needed only in revision)
            :type str
            @return: dict with the response and the attachment details
            """
            super(GetAttachmentOperation._GetAttachmentCommand, self).__init__(method="GET", is_read_request=True,
                                                                               use_stream=True)
            self._document_id = document_id
            self._name = name
            self._attachment_type = attachment_type
            self._change_vector = change_vector

        def create_request(self, server_node):
            self.url = "{0}/databases/{1}/attachments?id={2}&name={3}".format(
                server_node.url, server_node.database, Utils.quote_key(self._document_id),
                Utils.quote_key(self._name))

            if self._attachment_type != AttachmentType.document:
                self.method = "POST"
                self.data = {"Type": str(self._attachment_type), "ChangeVector": self._change_vector}

        def set_response(self, response):
            if response is None:
                return None

            if response.status_code == 200:
                attachment_details = {"content_type": response.headers.get("Content-Type", None),
                                      "change_vector": Utils.get_change_vector_from_header(response),
                                      "hash": response.headers.get("Attachment-Hash", None),
                                      "size": response.headers.get("Attachment-Size", 0)}

                return {"response": response, "details": attachment_details}


class PatchOperation(Operation):
    def __init__(self, document_id, change_vector, patch, patch_if_missing=None,
                 skip_patch_if_change_vector_mismatch=False):
        """
        @param str document_id: The id of the document
        @param str change_vector: The change_vector
        @param PatchRequest patch: The patch that going to be applied on the document
        @param PatchRequest patch_if_missing: The default patch to applied
        @param bool skip_patch_if_change_vector_mismatch: If True will skip documents that mismatch the change_vector
        """
        super(PatchOperation, self).__init__()
        self._document_id = document_id
        self._change_vector = change_vector
        self._patch = patch
        self._patch_if_missing = patch_if_missing
        self._skip_patch_if_change_vector_mismatch = skip_patch_if_change_vector_mismatch

    def get_command(self, store, conventions, cache=None):
        return PatchCommand(self._document_id, self._change_vector, self._patch, self._patch_if_missing,
                            self._skip_patch_if_change_vector_mismatch)


class PutAttachmentOperation(Operation):
    def __init__(self, document_id, name, stream, content_type=None, change_vector=None):
        """
        @param document_id: The id of the document
        @param name: Name of the attachment
        @param stream: The attachment as bytes (ex.open("file_path", "rb"))
        @param content_type: The type of the attachment (ex.image/png)
        @param change_vector: The change vector of the document
        """

        super(PutAttachmentOperation, self).__init__()
        self._document_id = document_id
        self._name = name
        self._stream = stream
        self._content_type = content_type
        self._change_vector = change_vector

    def get_command(self, store, conventions, cache=None):
        return PutAttachmentCommand(self._document_id, self._name, self._stream, self._content_type,
                                    self._change_vector)


class GetFacetsOperation(Operation):
    def __init__(self, query):
        """
        @param FacetQuery query: The query we wish to get
        """
        if query is None:
            raise ValueError("Invalid query")

        super(GetFacetsOperation, self).__init__()
        self._query = query

    def get_command(self, store, conventions, cache=None):
        return GetFacetsCommand(self._query)


class GetMultiFacetsOperation(Operation):
    def __init__(self, queries):
        if queries is None or len(queries) == 0:
            raise ValueError("Invalid queries")

        super(GetMultiFacetsOperation, self).__init__()
        self._queries = queries

    def get_command(self, store, conventions, cache=None):
        return self._GetMultiFacetsCommand(self._queries)

    class _GetMultiFacetsCommand(RavenCommand):
        def __init__(self, queries):
            super(GetMultiFacetsOperation._GetMultiFacetsCommand, self).__init__(is_read_request=True)
            requests = {}
            for q in queries:
                if not q:
                    raise ValueError("Invalid query")
                requests.update(
                    {"url": "/queries", "query": "?op=facets&query-hash=" + q.get_query_hash(), "method": "POST",
                     "data": q.to_json()})

            self._command = MultiGetCommand(requests)

        def create_request(self, server_node):
            return self._command.create_request(server_node)

        def set_response(self, response):
            results = self._command.set_response(response)
            return [result for result in results["Result"]]
